#! /usr/bin/env python
#
"""
Driver script for running the `housing` application on SMSCG.
"""
# Copyright (C) 2011, 2012  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.sjp
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>, Benjamin Jonen < benjamin.jonen@bf.uzh.ch>'
# summary of user-visible changes
__changelog__ = """
"""
__docformat__ = 'reStructuredText'


## Calls:
## -x /home/benjamin/workspace/fpProj/model/bin/forwardPremiumOut -b ../base/ para.loop  -C 1 -N -X i686
## 5-x /home/benjamin/workspace/idrisk/bin/idRiskOut -b ../base/ para.loop  -C 1 -N -X i686
## -x /home/benjamin/workspace/idrisk/model/bin/idRiskOut -b ../base/ para.loop  -C 1 -N
## Need to set path to linux kernel and apppot, e.g.: export PATH=$PATH:~/workspace/apppot:~/workspace/

# Preliminary imports
from __future__ import absolute_import, print_function
import os, shutil

## Remove all files in curPath if -N option specified.
if __name__ == '__main__':
    import sys
    if '-N' in sys.argv:

        path2Pymods = os.path.join(os.path.dirname(__file__), '../')
        if not sys.path.count(path2Pymods):
            sys.path.append(path2Pymods)
        curPath = os.getcwd()
        filesAndFolder = os.listdir(curPath)
        if 'ghousing.log' in filesAndFolder:
            if 'para.loop' in os.listdir(os.getcwd()):
                shutil.copyfile(os.path.join(curPath, 'para.loop'), os.path.join('/tmp', 'para.loop'))
                shutil.rmtree(curPath)
                shutil.copyfile(os.path.join('/tmp', 'para.loop'), os.path.join(curPath, 'para.loop'))
            else:
                shutil.rmtree(curPath)


# ugly workaround for Issue 95,
# see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import ghousing

# std module imports
import numpy as np
import re
import sys
import time
import cli

from supportGc3 import lower, flatten, str2tuple, getIndex, extractVal, str2vals
from supportGc3 import format_newVal, update_parameter_in_file, safe_eval, str2mat, mat2str, getParameter

from housing import housingApplication, housingApppotApplication
from paraLoop import paraLoop

path2Pymods = os.path.join(os.path.dirname(__file__), '../')
if not sys.path.count(path2Pymods):
    sys.path.append(path2Pymods)

from pymods.support.support import catFile
from pymods.support.support import wrapLogger
from pymods.classes.tableDict import tableDict

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.compat import lockfile
from gc3libs.cmdline import SessionBasedScript, existing_file
import gc3libs.utils
import gc3libs.application.apppot

#import gc3libs.debug

# import personal libraries
path2SrcPy = os.path.join(os.path.dirname(__file__), '../src')
if not sys.path.count(path2SrcPy):
    sys.path.append(path2SrcPy)
from plotSimulation import plotSimulation


### Temporary evil overloads

class Error(Exception):
    pass

class Abort(Error):
    """Raised when an application exits unexpectedly.

    :class:`Abort` takes a single integer argument indicating the exit status of
    the application.

    .. versionadded:: 1.0.4
    """

    def __init__(self, status):
        self.status = status
        message = "Application terminated (%s)" % self.status
        super(Abort, self).__init__(message, self.status)

def engineProgress(self):
    """
    Update state of all registered tasks and take appropriate action.
    Specifically:

      * tasks in `NEW` state are submitted;

      * the state of tasks in `SUBMITTED`, `RUNNING` or `STOPPED` state is updated;

      * when a task reaches `TERMINATING` state, its output is downloaded.

      * tasks in `TERMINATED` status are simply ignored.

    The `max_in_flight` and `max_submitted` limits (if >0) are
    taken into account when attempting submission of tasks.
    """
    # prepare
    currently_submitted = 0
    currently_in_flight = 0
    if self.max_in_flight > 0:
        limit_in_flight = self.max_in_flight
    else:
        limit_in_flight = utils.PlusInfinity()
    if self.max_submitted > 0:
        limit_submitted = self.max_submitted
    else:
        limit_submitted = utils.PlusInfinity()

    # update status of SUBMITTED/RUNNING tasks before launching
    # new ones, otherwise we would be checking the status of
    # some tasks twice...
    #gc3libs.log.debug("Engine.progress: updating status of tasks [%s]"
    #                  % str.join(', ', [str(task) for task in self._in_flight]))
    transitioned = []
    for index, task in enumerate(self._in_flight):
        try:
            self._core.update_job_state(task)
            if self._store and task.changed:
                self._store.save(task)
            state = task.execution.state
            if state == Run.State.SUBMITTED:
                # only real applications need to be counted
                # against the limit; policy tasks are exempt
                # (this applies to all similar clause below)
                if isinstance(task, Application):
                    currently_submitted += 1
                    currently_in_flight += 1
            elif state == Run.State.RUNNING:
                if isinstance(task, Application):
                    currently_in_flight += 1
            elif state == Run.State.STOPPED:
                transitioned.append(index) # task changed state, mark as to remove
                self._stopped.append(task)
            elif state == Run.State.TERMINATING:
                transitioned.append(index) # task changed state, mark as to remove
                self._terminating.append(task)
            elif state == Run.State.TERMINATED:
                transitioned.append(index) # task changed state, mark as to remove
                self._terminated.append(task)
        #except gc3libs.exceptions.ConfigurationError:
            ## Unrecoverable; no sense in continuing -- pass
            ## immediately on to client code and let it handle
            ## this...
            #raise
        except:
            gc3libs.log.debug('Error in updating task. Raising error. ')
            raise
    # remove tasks that transitioned to other states
    for index in reversed(transitioned):
        del self._in_flight[index]

    # execute kills and update count of submitted/in-flight tasks
    #gc3libs.log.debug("Engine.progress: killing tasks [%s]"
    #                  % str.join(', ', [str(task) for task in self._to_kill]))
    transitioned = []
    for index, task in enumerate(self._to_kill):
        try:
            old_state = task.execution.state
            self._core.kill(task)
            if self._store:
                self._store.save(task)
            if old_state == Run.State.SUBMITTED:
                if isinstance(task, Application):
                    currently_submitted -= 1
                    currently_in_flight -= 1
            elif old_state == Run.State.RUNNING:
                if isinstance(task, Application):
                    currently_in_flight -= 1
            self._terminated.append(task)
            transitioned.append(index)
        except Exception, x:
            gc3libs.log.error("Ignored error in killing task '%s': %s: %s"
                              % (task, x.__class__.__name__, str(x)),
                              exc_info=True)
    # remove tasks that transitioned to other states
    for index in reversed(transitioned):
        del self._to_kill[index]

    # update state of STOPPED tasks; again need to make before new
    # submissions, because it can alter the count of in-flight
    # tasks.
    #gc3libs.log.debug("Engine.progress: updating status of stopped tasks [%s]"
    #                  % str.join(', ', [str(task) for task in self._stopped]))
    transitioned = []
    for index, task in enumerate(self._stopped):
        try:
            self._core.update_job_state(task)
            if self._store and task.changed:
                self._store.save(task)
            state = task.execution.state
            if state in [Run.State.SUBMITTED, Run.State.RUNNING]:
                if isinstance(task, Application):
                    currently_in_flight += 1
                    if task.execution.state == Run.State.SUBMITTED:
                        currently_submitted += 1
                self._in_flight.append(task)
                transitioned.append(index) # task changed state, mark as to remove
            elif state == Run.State.TERMINATING:
                self._terminating.append(task)
                transitioned.append(index) # task changed state, mark as to remove
            elif state == Run.State.TERMINATED:
                self._terminated.append(task)
                transitioned.append(index) # task changed state, mark as to remove
        except Exception, x:
            gc3libs.log.error("Ignoring error in updating state of STOPPED task '%s': %s: %s"
                              % (task, x.__class__.__name__, str(x)),
                              exc_info=True)
        # except Exception, x:
        #     gc3libs.log.error("Ignoring error in updating state of STOPPED task '%s': %s: %s"
        #                       % (task, x.__class__.__name__, str(x)),
        #                       exc_info=True)
    # remove tasks that transitioned to other states
    for index in reversed(transitioned):
        del self._stopped[index]

    # now try to submit NEW tasks
    #gc3libs.log.debug("Engine.progress: submitting new tasks [%s]"
    #                  % str.join(', ', [str(task) for task in self._new]))
    transitioned = []
    if self.can_submit:
        index = 0
        while (currently_submitted < limit_submitted
               and currently_in_flight < limit_in_flight
               and index < len(self._new)):
            task = self._new[index]
            # try to submit; go to SUBMITTED if successful, FAILED if not
            if currently_submitted < limit_submitted and currently_in_flight < limit_in_flight:
                try:
                    self._core.submit(task)
                    if self._store:
                        self._store.save(task)
                    self._in_flight.append(task)
                    transitioned.append(index)
                    if isinstance(task, Application):
                        currently_submitted += 1
                        currently_in_flight += 1
                except Exception, x:
  #                  sys.excepthook(*sys.exc_info()) # DEBUG
                    import traceback
                    traceback.print_exc()
                    gc3libs.log.error("Ignored error in submitting task '%s': %s: %s"
                                      % (task, x.__class__.__name__, str(x)))
                    task.execution.history("Submission failed: %s: %s"
                                           % (x.__class__.__name__, str(x)))
                    #raise
                #except Exception:
                    #gc3libs.log.error("Ignored error in submitting task. BJ addition")
            index += 1
    # remove tasks that transitioned to SUBMITTED state
    for index in reversed(transitioned):
        del self._new[index]

    # finally, retrieve output of finished tasks
    #gc3libs.log.debug("Engine.progress: fetching output of tasks [%s]"
    #                  % str.join(', ', [str(task) for task in self._terminating]))
    if self.can_retrieve:
        transitioned = []
        for index, task in enumerate(self._terminating):
            # try to get output
            try:
                self._core.fetch_output(task)
            except gc3libs.exceptions.UnrecoverableDataStagingError, ex:
                gc3libs.log.error("Error in fetching output of task '%s',"
                                  " will mark it as TERMINATED"
                                  " (with error exit code %d): %s: %s",
                                  task, posix.EX_IOERR,
                                  ex.__class__.__name__, str(ex), exc_info=True)
                task.execution.returncode = (Run.Signals.DataStagingFailure,
                                             posix.EX_IOERR)
                task.execution.state = Run.State.TERMINATED
                task.changed = True
            #except Exception, x:
                #gc3libs.log.error("Ignored error in fetching output of task '%s': %s: %s"
                                  #% (task, x.__class__.__name__, str(x)), exc_info=True)
            if task.execution.state == Run.State.TERMINATED:
                self._terminated.append(task)
                self._core.free(task)
                transitioned.append(index)
            if self._store and task.changed:
                self._store.save(task)
        # remove tasks for which final output has been retrieved
        for index in reversed(transitioned):
            del self._terminating[index]

import gc3libs.core
gc3libs.core.Engine.progress = engineProgress



def pre_run(self):
    """
        Temporary overload for pre_run method of gc3libs.cmdline._Script.
    """
    import cli # pyCLI
    import cli.app
    import cli._ext.argparse as argparse
    from cli.util import ifelse, ismethodof
    import logging
    ## finish setup
    self.setup_options()
    self.setup_args()

    ## parse command-line
    cli.app.CommandLineApp.pre_run(self)

    ## setup GC3Libs logging
    loglevel = max(1, logging.ERROR - 10 * max(0, self.params.verbose - self.verbose_logging_threshold))
    gc3libs.configure_logger(loglevel, self.name)
    self.log = logging.getLogger('gc3.gc3utils') # alternate: ('gc3.' + self.name)
    self.log.setLevel(loglevel)

    self.log.propagate = True
    self.log.parent.propagate = False
    # Changed to false since we want to avoid dealing with the root logger and catch the information directly.

    from logging import getLogger
    from logbook.compat import redirect_logging
    from logbook.compat import RedirectLoggingHandler
#    redirect_logging() # does the same thing as adding a RedirectLoggingHandler... might as well be explicit
    self.log.parent.handlers = []
    self.log.parent.addHandler(RedirectLoggingHandler())
    print self.log.handlers
    print self.log.parent.handlers
    print self.log.root.handlers

    self.log.critical('Successfully overridden gc3pie error handling. ')

    # interface to the GC3Libs main functionality
    self._core = self._make_core()

    # call hook methods from derived classes
    self.parse_args()

logger = wrapLogger(loggerName = 'ghousing.log', streamVerb = 'INFO', logFile = os.path.join(os.getcwd(), 'ghousing.log'))
gc3utilsLogger = wrapLogger(loggerName = 'gc3ghousing.log', streamVerb = 'INFO', logFile = os.path.join(os.getcwd(), 'ghousing.log'),
                            streamFormat = '{record.time:%Y-%m-%d %H:%M:%S} - {record.channel}: {record.message}',
                            fileFormat = '{record.time:%Y-%m-%d %H:%M:%S} - {record.channel}: {record.message}')

def dispatch_record(record):
    """Passes a record on to the handlers on the stack.  This is useful when
    log records are created programmatically and already have all the
    information attached and should be dispatched independent of a logger.
    """
#    logbook.base._default_dispatcher.call_handlers(record)
    gc3utilsLogger.call_handlers(record)

import gc3libs.cmdline
#gc3libs.cmdline._Script.pre_run = pre_run
import logbook
#logbook.dispatch_record = dispatch_record


## custom application class


class ghousing(SessionBasedScript, paraLoop):
    """
Read `.loop` files and execute the `housingOut` program accordingly.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = '1.0',
            input_filename_pattern = '*.loop',
            stats_only_for = Application,
        )
        paraLoop.__init__(self, 'INFO')

    def setup_options(self):
        self.add_param("-b", "--initial", metavar="DIR",
                       dest="initial",
                       help="Include directory contents in any job's input."
                       " Use this to specify the initial guess files.")
        self.add_param("-n", "--dry-run",
                       dest = 'dryrun', action="store_true", default = False,
                       help="Take the loop for a test run")
        self.add_param("-x", "--executable", metavar="PATH",
                       dest="executable", default=os.path.join(
                           os.getcwd(), "forwardPremiumOut"),
                       help="Path to the `idRisk` executable binary"
                       "(Default: %(default)s)")
        self.add_param("-X", "--architecture", metavar="ARCH",
                       dest="architecture", default=Run.Arch.X86_64,
                       help="Processor architecture required by the executable"
                       " (one of: 'i686' or 'x86_64', without quotes)")
        self.add_param("-A", "--apppot", metavar="PATH",
                       dest="apppot",
                       type=existing_file, default=None,
                       help="Use an AppPot image to run idRisk."
                       " PATH can point either to a complete AppPot system image"
                       " file, or to a `.changes` file generated with the"
                       " `apppot-snap` utility.")
        self.add_param("-rte", "--rte", action="store_true", default=None,
                       help="Use an AppPot image to run idRisk."
                       "use predeployed image")




    def parse_args(self):
        """
        Check validity and consistency of command-line options.
        """
        if not os.path.exists(self.params.executable):
            raise gc3libs.exceptions.InvalidUsage(
                "Path '%s' to the 'housingOut' executable does not exist;"
                " use the '-x' option to specify a valid one."
                % self.params.executable)
        if os.path.isdir(self.params.executable):
            self.params.executable = os.path.join(self.params.executable,
                                                  'housing')
        gc3libs.utils.check_file_access(self.params.executable, os.R_OK|os.X_OK,
                                gc3libs.exceptions.InvalidUsage)

    def run(self):
        """
        Execute `cli.app.Application.run`:meth: if any exception is
        raised, catch it, output an error message and then exit with
        an appropriate error code.
        """
        try:
            return cli.app.CommandLineApp.run(self)
        except gc3libs.exceptions.InvalidUsage, ex:
            # Fatal errors do their own printing, we only add a short usage message
            sys.stderr.write("Type '%s --help' to get usage help.\n" % self.name)
            return 64  # EX_USAGE in /usr/include/sysexits.h
        except KeyboardInterrupt:
            sys.stderr.write("%s: Exiting upon user request (Ctrl+C)\n" % self.name)
            return 13
        except SystemExit, ex:
            return ex.code
        # the following exception handlers put their error message
        # into `msg` and the exit code into `rc`; the closing stanza
        # tries to log the message and only outputs it to stderr if
        # this fails
        except lockfile.Error, ex:
            exc_info = sys.exc_info()
            msg = ("Error manipulating the lock file (%s: %s)."
                   " This likely points to a filesystem error"
                   " or a stale process holding the lock."
                   " If you cannot get this command to run after"
                   " a system reboot, please write to gc3pie@googlegroups.com"
                   " including any output you got by running '%s -vvvv %s'.")
            if len(sys.argv) > 0:
                msg %= (ex.__class__.__name__, str(ex),
                        self.name, str.join(' ', sys.argv[1:]))
            else:
                msg %= (ex.__class__.__name__, str(ex), self.name, '')
            rc = 1
        except AssertionError, ex:
            exc_info = sys.exc_info()
            msg = ("BUG: %s\n"
                   "Please send an email to gc3pie@googlegroups.com"
                   " including any output you got by running '%s -vvvv %s'."
                   " Thanks for your cooperation!")
            if len(sys.argv) > 0:
                msg %= (str(ex), self.name, str.join(' ', sys.argv[1:]))
            else:
                msg %= (str(ex), self.name, '')
            rc = 1
        #except Exception, ex:
            #msg = "%s: %s" % (ex.__class__.__name__, str(ex))
            #if isinstance(ex, cli.app.Abort):
                #rc = (ex.status)
            #elif isinstance(ex, EnvironmentError):
                #rc = 74  # EX_IOERR in /usr/include/sysexits.h
            #else:
                ## generic error exit
                #rc = 1
        ## output error message and -maybe- backtrace...
        #try:
            #self.log.critical(msg,
                              #exc_info=(self.params.verbose > self.verbose_logging_threshold + 2))
        #except:
            ## no logging setup, output to stderr
            #sys.stderr.write("%s: FATAL ERROR: %s\n" % (self.name, msg))
            #if self.params.verbose > self.verbose_logging_threshold + 2:
                #sys.excepthook(* sys.exc_info())
        ## ...and exit
        #return 1


    #def run(self):
        #"""
        #Execute `cli.app.Application.run`:meth: if any exception is
        #raised, catch it, output an error message and then exit with
        #an appropriate error code.
        #"""

      ##  return cli.app.CommandLineApp.run(self)
        #import cli
        #try:
            #return cli.app.CommandLineApp.run(self)
        #except gc3libs.exceptions.InvalidUsage, ex:
            ## Fatal errors do their own printing, we only add a short usage message
            #sys.stderr.write("Type '%s --help' to get usage help.\n" % self.name)
            #return 64 # EX_USAGE in /usr/include/sysexits.h
        #except KeyboardInterrupt:
            #sys.stderr.write("%s: Exiting upon user request (Ctrl+C)\n" % self.name)
            #return 13
        #except SystemExit, ex:
            #return ex.code
        ## the following exception handlers put their error message
        ## into `msg` and the exit code into `rc`; the closing stanza
        ## tries to log the message and only outputs it to stderr if
        ## this fails
        #except lockfile.Error, ex:
            #exc_info = sys.exc_info()
            #msg = ("Error manipulating the lock file (%s: %s)."
                   #" This likely points to a filesystem error"
                   #" or a stale process holding the lock."
                   #" If you cannot get this command to run after"
                   #" a system reboot, please write to gc3pie@googlegroups.com"
                   #" including any output you got by running '%s -vvvv %s'.")
            #if len(sys.argv) > 0:
                #msg %= (ex.__class__.__name__, str(ex),
                        #self.name, str.join(' ', sys.argv[1:]))
            #else:
                #msg %= (ex.__class__.__name__, str(ex), self.name, '')
            #rc = 1
        #except AssertionError, ex:
            #exc_info = sys.exc_info()
            #msg = ("BUG: %s\n"
                   #"Please send an email to gc3pie@googlegroups.com"
                   #" including any output you got by running '%s -vvvv %s'."
                   #" Thanks for your cooperation!")
            #if len(sys.argv) > 0:
                #msg %= (str(ex), self.name, str.join(' ', sys.argv[1:]))
            #else:
                #msg %= (str(ex), self.name, '')
            #rc = 1
        #except NotImplementedError, nE:
            #pass

    def new_tasks(self, extra):
        # Generating new tasks for both paramter files
        self.logger.info('\ngenParamters.in: ')
        self.logger.info('-----------------')
        catFile(os.path.join(self.params.initial, 'input', 'genParameters.in'))
        self.logger.info('\nctryParameters.in: ')
        self.logger.info('-----------------')
        catFile(os.path.join(self.params.initial, 'input', 'ctryParameters.in'))
        self.logger.info('\npara.loop: ')
        self.logger.info('-----------------')
        catFile(os.path.join(os.getcwd(), 'para.loop'))
        self.logger.info('\nPress [y] to confirm the input files and continue execution of ghousing.py. Press [q] to exit')
        #import select
        #rlist, wlist, xlist = select.select([sys.stdin], [], [], None)
        selection = raw_input()
        if selection.lower() == 'q':
            self.logger.critical('Exiting upon user request...')
            os._exit(1)
        # setup AppPot parameters
        use_apppot = False
        #bug = use_apppot[0]
        apppot_img = None
        apppot_changes = None
        if self.params.apppot:
            use_apppot = True
            if self.params.apppot.endswith('.changes.tar.gz'):
                apppot_changes = self.params.apppot
            else:
                apppot_img = self.params.apppot

        inputs = self._search_for_input_files(self.params.args)

        # create a tar.gz archive of the code
        import tarfile
        tar = tarfile.open(os.path.join(os.getcwd(), 'codeBase.tar.gz'), "w:gz")
        for name in [self.params.initial, os.path.join(self.params.initial, '../code')]:
            tar.add(name)
        tar.close()

        # Copy base dir
        localBaseDir = os.path.join(os.getcwd(), 'localBaseDir')
        shutil.copytree(self.params.initial, localBaseDir)

        # update ctry Parameters. Important, before I do the para.loop adjustments
        ctryInParaLoop = False
        for para_loop in inputs:
            if os.path.isdir(para_loop):
                para_loop = os.path.join(para_loop, 'para.loop')
            paraLoopFile = open(para_loop, 'r')
            paraLoopFile.readline()
            for line in paraLoopFile:
                if not line.rstrip():
                    continue
                eles = line.split()
                var = eles[0]
                val = eles[6]
                if var == 'ctry':
                    ctryInParaLoop = True
                    ctry = val
        localBaseDirInputFolder = os.path.join(localBaseDir, 'input')
        genParametersFile = os.path.join(localBaseDirInputFolder, 'genParameters.in')
        ctryParametersFile = os.path.join(localBaseDirInputFolder, 'ctryParameters.in')
        updateCtryParametersFile = bool(getParameter(genParametersFile, 'updateCtryParametersFile', 'space-separated'))
        if not ctryInParaLoop:
            ctry = getParameter(genParametersFile, 'ctry', 'space-separated')
        if updateCtryParametersFile:
            shutil.copy(os.path.join(localBaseDirInputFolder, ctry + 'CtryParameters.in'), os.path.join(localBaseDirInputFolder, 'ctryParameters.in'))

        for para_loop in inputs:
            path_to_base_dir = os.path.dirname(para_loop)
#            self.log.debug("Processing loop file '%s' ...", para_loop)
            for jobname, substs in self.process_para_file(para_loop):
##                self.log.debug("Job '%s' defined by substitutions: %s.",
##                               jobname, substs)
                executable = os.path.basename(self.params.executable)
                inputs = { self.params.executable:executable }
                # make a "stage" directory where input files are collected
                path_to_stage_dir = self.make_directory_path(self.params.output, jobname)
                input_dir = path_to_stage_dir #os.path.join(path_to_stage_dir, 'input')
                gc3libs.utils.mkdir(input_dir)
                prefix_len = len(input_dir) + 1
                # 2. apply substitutions to parameter files
                for (path, changes) in substs.iteritems():
                    for (var, val, index, regex) in changes:
                        update_parameter_in_file(os.path.join(localBaseDir, path),
                                                 var, index, val, regex)
                fillInputDir(localBaseDir, input_dir)
                # 3. build input file list
                for dirpath,dirnames,filenames in os.walk(input_dir):
                    for filename in filenames:
                        # cut the leading part, which is == to path_to_stage_dir
                        relpath = dirpath[prefix_len:]
                        # ignore output directory contents in resubmission
                        if relpath. startswith('output'):
                            continue
                        remote_path = os.path.join(relpath, filename)
                        inputs[os.path.join(dirpath, filename)] = remote_path
                # all contents of the `output` directory are to be fetched
                outputs = { 'output/':'' }
                kwargs = extra.copy()
                kwargs['stdout'] = os.path.join('housingStdOut.log')
                kwargs['join'] = True
                kwargs['output_dir'] = os.path.join(path_to_stage_dir, 'output')
                kwargs['requested_architecture'] = self.params.architecture

#                print 'inputs = %s' % inputs
#                print 'outputs = %s' % outputs

#                kwargs.setdefault('tags', [ ])

                # adaptions for uml
                if self.params.rte:
                    kwargs['apppot_tag'] = 'ENV/APPPOT-0.26'
#                    kwargs['tags'] = ['TEST/APPPOT-IBF-1.0']
#                    kwargs['tags'] = ['TEST/APPPOT-IBF-1.1']
                    kwargs['tags'] = ['APPS/ECON/APPPOT-IBF-1.0']

                    cls = housingApppotApplication
                    pathToExecutable = '/home/user/job/' + executable
                elif use_apppot:
                    if apppot_img is not None:
                        kwargs['apppot_img'] = apppot_img
                    if apppot_changes is not None:
                        kwargs['apppot_changes'] = apppot_changes
                    cls = housingApppotApplication
                    pathToExecutable = '/home/user/job/' + executable
                else:
                    cls = housingApplication
                    pathToExecutable = executable

#                print 'kwargs = %s' % kwargs
                # hand over job to create
                yield (jobname, cls, [pathToExecutable, [], inputs, outputs], kwargs)

def fillInputDir(baseDir, input_dir):
    '''
      Copy folder /input and all files in the base dir to input_dir.
      This is slightly more involved than before because we need to
      exclude the markov directory which contains markov information
      for all country pairs.
    '''
    shutil.copytree(baseDir, input_dir)

def combinedThresholdPlot():
    import copy
    folders = [folder for folder in os.listdir(os.getcwd()) if os.path.isdir(folder) and not folder == 'localBaseDir' and not folder == 'ghousing.jobs']
    tableList = [ (folder, os.path.join(os.getcwd(), folder, 'output', 'ownershipThreshold_1.out')) for folder in folders ]
    tableDicts = dict([ (folder, tableDict.fromTextFile(table, width = np.max([len(folder) for folder in folders]) + 5, prec = 10)) for folder, table in tableList if os.path.isfile(table)])
    tableKeys = tableDicts.keys()
    tableKeys.sort()
    if tableDicts:
        for ixTable, tableKey in enumerate(tableKeys):
#            print tableKey
#            print tableDicts[tableKey]
            table = copy.deepcopy(tableDicts[tableKey])
            table.keep(['age', 'yst1'])
            table.rename('yst1', tableKey)
            if ixTable == 0:
                fullTable = copy.deepcopy(table)
            else:
                fullTable.merge(table, 'age')
                if '_merge' in fullTable.cols:
                    fullTable.drop('_merge')
                logger.info(fullTable)
        logger.info(fullTable)
        f = open(os.path.join(os.getcwd(), 'ownerThresholds'), 'w')
        print >> f, fullTable
        f.flush()
        ax = 3
        plotSimulation(table = os.path.join(os.getcwd(), 'ownerThresholds'), xVar = 'age', yVars = list(fullTable.cols), figureFile = os.path.join(os.getcwd(), 'ownerThresholds.png'), verb = 'CRITICAL' )

def combinedOwnerSimuPlot():
    import copy
    logger.debug('starting combinedOwnerSimuPlot')
    folders = [folder for folder in os.listdir(os.getcwd()) if os.path.isdir(folder) and not folder == 'localBaseDir' and not folder == 'ghousing.jobs']
    logger.debug('folders are %s ' % folders)
    tableList = [ (folder, os.path.join(os.getcwd(), folder, 'output', 'aggregate.out')) for folder in folders ]
    tableDicts = dict([ (folder, tableDict.fromTextFile(table, width = np.max([len(folder) for folder in folders]) + 5, prec = 10)) for folder, table in tableList if os.path.isfile(table)])
    tableKeys = tableDicts.keys()
    tableKeys.sort()
    if tableDicts:
        for ixTable, tableKey in enumerate(tableKeys):
            table = copy.deepcopy(tableDicts[tableKey])
            table.keep(['age', 'owner'])
            table.rename('owner', tableKey)
            if ixTable == 0:
                fullTable = copy.deepcopy(table)
            else:
                fullTable.merge(table, 'age')
                fullTable.drop('_merge')
            logger.info(fullTable)

        empOwnershipFile = os.path.join(os.getcwd(), 'localBaseDir', 'input', 'PSIDOwnershipProfilealleduc.out')
        empOwnershipTable = tableDict.fromTextFile(empOwnershipFile, width = 20, prec = 10)
        empOwnershipTable.rename('PrOwnership', 'empOwnership')
        fullTable.merge(empOwnershipTable, 'age')
        fullTable.drop('_merge')
        if fullTable:
            logger.info(fullTable)
            f = open(os.path.join(os.getcwd(), 'ownerSimu'), 'w')
            print >> f, fullTable
            f.flush()
        else:
            logger.info('no owner simus')
        logger.debug('done combinedOwnerSimuPlot')

        plotSimulation(table = os.path.join(os.getcwd(), 'ownerSimu'), xVar = 'age', yVars = list(fullTable.cols), yVarRange = (0., 1.), figureFile = os.path.join(os.getcwd(), 'ownerSimu.png'), verb = 'CRITICAL' )

def combineRunningTimes():
    folders = [folder for folder in os.listdir(os.getcwd()) if os.path.isdir(folder) and not folder == 'localBaseDir' and not folder == 'ghousing.jobs']
    runTimeFileList = [ (folder, os.path.join(os.getcwd(), folder, 'output', 'runningTime.out')) for folder in folders ]
    print runTimeFileList
    runTimes = {} # in minutes
    for folder, fle in runTimeFileList:
        if not os.path.isfile(fle): continue
        runningTimeFile = open(fle)
        lines = runningTimeFile.readlines()
        for line in lines:
            if line.find('Full running'):
                match = re.match('(.*=)([0-9\.\s]*)(.*)', line.rstrip()).groups()
                if not match:
                    runTimeSec = 0.
                else:
                    runTimeSec = float(match [1].strip()) # in seconds
        runTimes[folder] = runTimeSec / 60.
    logger.info('running times are \n %s' % runTimes)
    f = open(os.path.join(os.getcwd(), 'runTimes.out'), 'w')
    folderKeys = runTimes.keys()
    folderKeys.sort()
    for key in folderKeys:
        print >> f, '%s = %f12.1' % (key, runTimes[key])
    f.flush()
    f.close()

def getDateTimeStr():
    import datetime
    cDate = datetime.date.today()
    cTime = datetime.datetime.time(datetime.datetime.now())
    dateString = '%04d-%02d-%02d-%02d-%02d-%02d' % (cDate.year, cDate.month, cDate.day, cTime.hour, cTime.minute, cTime.second)
    return dateString


## run scriptfg


if __name__ == '__main__':
    #logger.info('Starting: \n%s' % ' '.join(sys.argv))
    logger.info('\n%s - Starting: \n%s' % (getDateTimeStr(), ' '.join(sys.argv)))
    os.system('cat ghousing.log')
    os.system('cd ~/workspace/housingProj/model/code && hg log > ' + os.path.join(os.getcwd(), 'hgLog'))
    #os._exit(1)
    ghousing().run()
    #from guppy import hpy
    #h = hpy()
    #print h.heap()
    # create overview plots across parameter combinations
    try:
        combinedThresholdPlot()
    except:
        logger.critical('problem creating combinedThresholdPlot. Investigate...')
    try:
        combinedOwnerSimuPlot()
    except:
        logger.critical('problem creating combinedOwnerSimuPlot. Investigate...')
    try:
        combineRunningTimes()
    except:
        logger.critical('problem creating combineRunningTimes. Investigate...')
    # some find commands to copy result graphs to a common directory.
    os.system("find -maxdepth 1 -type d -iregex './p.*' -exec bash -c 'x='{}' && mkdir -p ownerPlots && y=${x#./} && echo $y && cp ${y}/output/ownership_aggregate.out.png ownerPlots/${y}.png 2>/dev/null' \; ; ld ownerPlots/")
    os.system("find -maxdepth 1 -type d -iregex './p.*' -exec bash -c 'x='{}' && mkdir -p aggregatePlots && y=${x#./} && echo $y && cp ${y}/output/aggregate.png aggregatePlots/${y}.png' \; ; ld aggregatePlots/")
    os.system("find -maxdepth 1 -type d -iregex './p.*' -exec bash -c 'x='{}' && mkdir -p ownerBdry && y=${x#./} && echo $y && cp ${y}/output/ownerBdry1.png ownerBdry/${y}.png' \; ; ld ownerBdry")

    logger.info('main done')



#! /usr/bin/env python
#
#   ggatk.py -- Front-end script for running ParRecoveryFun Matlab
#   function with a given combination of reference models.
#
#   Copyright (C) 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2015-11-17:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import ggatk
    ggatk.GgatkScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
import random
import posix

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, \
    hours, minutes, seconds
from gc3libs.workflow import RetryableTask, StagedTaskCollection, \
    ParallelTaskCollection, SequentialTaskCollection

gatk_steps_list = ['0','1','2']

# Utility methods
def get_bams(input_folder):
    """
    Return list of .bam .bam.bai pairs
    """
    for bam in [ bams for bams in os.listdir(input_folder) if bams.endswith('.bam')
                 and os.path.isfile(os.path.join(input_folder,bams+".bai"))]:
        yield (os.path.join(input_folder,bam),
               os.path.join(input_folder,bam+".bai"))

def get_vcf_group(input_folder, group_size):
    """
    Return list of .vcf file lists of size `group_size`

    """
    vcf_list = [ os.path.join(input_folder,vcfs) for vcfs in os.listdir(input_folder) \
                 if vcfs.endswith(".vcf") ]
    for i in xrange(0, len(vcf_list), group_size):
        yield (vcf_list[i:i+group_size],i)

## custom application class
class GATKS0Application(Application):
    """
    GATK Stage0: run GATK steps 1,2,3 on a given .bam (and .bai) file
    """
    application_name = 'gatks0'

    def __init__(self, bam_file, bai_file, **extra_args):

        self.S0_output = extra_args['S0_output']

        extra_args['requested_memory'] = extra_args['S0_memory']

        inputs = dict()
        outputs = dict()
        executables = []

        # execution wrapper needs to be added anyway
        gatks0_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gatks0.sh")
        inputs[gatks0_wrapper_sh] = os.path.basename(gatks0_wrapper_sh)

        arguments = "./%s %s %s %s -m %s" % (os.path.basename(gatks0_wrapper_sh),
                                             extra_args['bam_filename'],
                                             extra_args['bai_filename'],
                                             extra_args['sample_name'],
                                             str(extra_args['requested_memory'].amount(conv=int)))

        inputs[bam_file] = extra_args['bam_filename']
        inputs[bai_file] = extra_args['bai_filename']

        # Set output
        self.vcf_output_filename = "./%s.g.vcf" % extra_args['sample_name']
        self.vcfindx_output_filename = "./%s.g.vcf.idx" % extra_args['sample_name']
        outputs[self.vcf_output_filename] = os.path.join(
            self.S0_output,
            self.vcf_output_filename)

        outputs[self.vcfindx_output_filename] = os.path.join(
            self.S0_output,
            self.vcfindx_output_filename)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gatks0.log',
            join=True,
            executables = os.path.basename(gatks0_wrapper_sh),
            **extra_args)

    def terminated(self):
        """
        Verify output
        """
        vcf_output = os.path.join(self.S0_output, self.vcf_output_filename)
        vcf_indx_output = os.path.join(self.S0_output, self.vcfindx_output_filename)
        try:
            assert os.path.isfile(vcf_output), \
                "Output file %s not found." % vcf_output

            assert os.path.isfile(vcf_indx_output), \
                "Output file %s not found." % vcf_indx_output

        except AssertionError as ex:
            self.execution.returncode = (0, posix.EX_OSFILE)
            gc3libs.log.error(ex.message)

class GATKS1Application(Application):
    """
    GATK Stage1: run GATK CombineGVCFs step
    XXX: does the corresponding .vcf.indx file has to be included ? YES
    """
    application_name = 'gatks1'

    def __init__(self, vcf_group, index, **extra_args):

        self.S1_output = extra_args['S1_output']
        extra_args['requested_memory'] = extra_args['S1_memory']

        inputs = dict()
        outputs = dict()
        executables = []

        # execution wrapper needs to be added anyway
        gatks1_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gatks1.sh")
        inputs[gatks1_wrapper_sh] = os.path.basename(gatks1_wrapper_sh)

        arguments = "./%s ./input %s -m %s" % (os.path.basename(gatks1_wrapper_sh),
                                               index,
                                               str(extra_args['requested_memory'].amount(conv=int)))

        for vcf in vcf_group:
            inputs[vcf] = os.path.join('./input',os.path.basename(vcf))

        # Set output
        self.vcf_output_filename = "./combined%d.g.vcf" % index
        outputs[self.vcf_output_filename] = os.path.join(
            self.S1_output,
            self.vcf_output_filename)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gatks1.log',
            join=True,
            executables = os.path.basename(gatks1_wrapper_sh),
            **extra_args)

    def terminated(self):
        """
        Verify output
        """
        vcf_output = os.path.join(self.S1_output, self.vcf_output_filename)
        try:
            assert os.path.isfile(vcf_output), \
                "Output file %s not found." % vcf_output
        except AssertionError as ex:
            self.execution.returncode = (0, posix.EX_OSFILE)
            gc3libs.log.ERROR(ex.message)

class GATKS2Application(Application):
    """
    GATK Stage2: run GATK CombineGVCFs step
    XXX: does the corresponding .vcf.indx file has to be included ?
    """
    application_name = 'gatks2'

    def __init__(self, vcf_list, **extra_args):

        self.S2_output = extra_args['S2_output']
        extra_args['requested_memory'] = extra_args['S2_memory']

        inputs = dict()
        outputs = dict()
        executables = []

        # execution wrapper needs to be added anyway
        gatks2_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gatks2.sh")
        inputs[gatks2_wrapper_sh] = os.path.basename(gatks2_wrapper_sh)

        arguments = "./%s ./input -m %s " % (os.path.basename(gatks2_wrapper_sh),
                                             str(extra_args['requested_memory'].amount(conv=int)))

        for vcf in vcf_list:
            inputs[vcf] = os.path.join('./input',os.path.basename(vcf))

        # Set output
        self.vcf_output = os.path.join(self.S2_output,
                                       'genotyped.gvcf.vcf')
        outputs['genotyped.gvcf.vcf'] = self.vcf_output

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gatks2.log',
            join=True,
            executables = os.path.basename(gatks2_wrapper_sh),
            **extra_args)

    def terminated(self):
        """
        Verify output
        """
        try:
            assert os.path.isfile(self.vcf_output), \
                "Output file %s not found." % vcf_output
        except AssertionError as ex:
            self.execution.returncode = (0, posix.EX_OSFILE)
            gc3libs.log.ERROR(ex.message)

# class GATKStagedTaskCollection(StagedTaskCollection):
#     """
#     Stage0: for each sample run GATK pipeline steps 1,2,3
#     * 1 sample takes 24-72 hours on single core
#     * GATK can be scripted to run individual steps
#     * Output: 2 files per sample (g.vcf and g.vcf.idx size 1GB total)
#     Stage1: Run GATK pipeline steps 4,5 on all g.vcf files together
#     * walltime and memory requirement unknown
#     * steps 4,5 can also be run on groups of g.vcf files but then
#     results need to be combined (GATK step 6)
#     * We will see whether this step is necessary resource-wise
#     * Output: 1 .vcf file size: 50GB
#     Stage2: Filter .vcf file
#     * Run vcftools and own scripts
#     * Output: single .vcf file size: 50MB
#     """
#     def __init__(self, input_bam_folder, **extra_args):
#         self.S0_output = extra_args['S0_output_dir']
#         self.S1_output = extra_args['S1_output_dir']
#         self.input_bam_folder = input_bam_folder
#         self.name = os.path.basename(input_bam_folder)
#         self.extra = extra_args
#         self.output_dir = extra_args['output_dir']
#         StagedTaskCollection.__init__(self)

class GATKSequentialTaskCollection(SequentialTaskCollection):
    def __init__(self, input_bam_folder, **extra_args):
        self.gatkseq = extra_args['gatk_sequence']
        self.S0_output = extra_args['S0_output']
        self.S1_output = extra_args['S1_output']
        self.input_bam_folder = input_bam_folder
        self.name = os.path.basename(input_bam_folder)
        self.extra = extra_args
        self.output_dir = extra_args['output_dir']

        self.jobname = "GATK-" + "_".join(self.gatkseq)

        # Get the first stage task in gatk_sequence
        initial_task_fn = getattr(self, "stage%d" % int(self.gatkseq.pop(0)))
        initial_task = initial_task_fn()

        SequentialTaskCollection.__init__(self, [initial_task])

    def next(self,done):
        """
        self.tasks[done]
        """
        if self.gatkseq:
            try:
                next_gatk = int(self.gatkseq.pop(0))
                next_stage_fn = getattr(self, "stage%d" % next_gatk)
            except AttributeError:
                gc3libs.log.debug("GATK sequence '%s' has no stage%d,"
                                  " ending sequence now.", self, next_gatk)
                self.execution.returncode = self.tasks[done].execution.returncode
                return Run.State.TERMINATED

            # get next stage (2); if we get an error here, something is wrong in
            # the code
            try:
                next_stage = next_stage_fn()
            except AttributeError as err:
                raise AssertionError(
                    "Invalid `Task` instance %r: %s"
                    % (self, str(err)))
            # add next stage to the collection, or end graciously
            if isinstance(next_stage, Task):
                self.add(next_stage)
                return Run.State.RUNNING
            elif isinstance(next_stage, (int, long, tuple)):
                self.execution.returncode = next_stage
                return Run.State.TERMINATED
            else:
                raise AssertionError(
                    "Invalid return value from method `stage%d()` of"
                    " `StagedTaskCollection` object %r: must return `Task`"
                    " instance or number" % (done + 1, self))
        else:
            self.execution.returncode = self.tasks[done].execution.returncode
            return Run.State.TERMINATED


    def stage0(self):
        """
        Stage0: for each sample run GATK pipeline steps 1,2,3
        * 1 sample takes 24-72 hours on single core
        * GATK can be scripted to run individual steps
        * Output: 2 files per sample (g.vcf and g.vcf.idx size 1GB total)
        # 300 samples - see if we can allocate 150 cores for 2 days
        # 1 day each
        Example script:
java -jar -d64 ~/programs/GenomeAnalysisTK.jar\
     -T HaplotypeCaller\
     --emitRefConfidence GVCF\
     -minPruning 3 -stand_call_conf 30 \
     -stand_emit_conf 10 \
     -R ~/goat.genome/goat_scaffoldFG_V1.1.normalised.22.07.fa -I \
        $file -o ${samplename}.g.vcf
        """

        tasks = []

        for (bam_file,bai_file) in get_bams(self.input_bam_folder):
            extra_args = self.extra.copy()
            extra_args['sample_name'] = os.path.basename(bam_file).split('.')[0]
            extra_args['bam_filename'] = os.path.basename(bam_file)
            extra_args['bai_filename'] = os.path.basename(bai_file)
            extra_args['jobname'] = "gatk-s0-%s" % extra_args['bam_filename']

            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        extra_args['jobname'])

            gc3libs.log.debug("Creating Stage0 task for : %s" %
                              (extra_args['bam_filename']))

            tasks.append(GATKS0Application(
                bam_file,
                bai_file,
                **extra_args))

        return ParallelTaskCollection(tasks)


    def stage1(self):
        """
        Start this stage IIF stage0 all completed (i.e. no failures)
        combine all .g.vcf files alltogether
        group in blocks (e.g. 30 out of the total 300)
        * make grouping an option for stage1
        * Use same GATK and goat.genome vesion as in stage0
        Run "combine_gvcf" script
        script can take an arbitrary number of gvc files and prodices
        1 single gvcf file
        end of stage1: 10 .g.vcf files
        if fails - because of heap size - then re-run with more memory
        Walltime: 2days each
        Cores requires: 10 cores
        Memory 500GB memory top - need to check
        memory: 128GB
        Example script:
java -jar  /home/dleigh/GenomeAnalysisTK-3.1-1/GenomeAnalysisTK-3.4-46/GenomeAnalysisTK.jar \
    -T CombineGVCFs \
    -R /home/dleigh/goatgenome/01.GENOME/scaffold/goat_scaffoldFG_V1.1.normalised.22.07.fa \
--variant /home/dleigh/demultiplexed.reads/GATK/GR0766.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1380.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1387.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1390.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1422.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1424.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1440.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1441.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1709.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1728.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1938.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1939.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR1997.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR2001.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR2053.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR2055.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/GR2056.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/SG0038.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/SG0047.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/SG0101.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/SG0242.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/SG0258.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/SG0261.g.vcf \
--variant /home/dleigh/demultiplexed.reads/GATK/SG0306.g.vcf \
-o /home/dleigh/demultiplexed.reads/GATK/combined3.g.vcf

        get list of all outputs in 'outputs0' folder
        group them in 's1_chunk'
        for each group run GATKS1Application
        """
        # XXX: add check if stage0 completed properly
        # Stop otherwise

        tasks = []

        for (vcf_group,index) in get_vcf_group(self.extra['S0_output'],
                                               int(self.extra['S1_group'])):
            extra_args = self.extra.copy()
            extra_args['jobname'] = "gatk-s1-%d" % index

            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        extra_args['jobname'])

            gc3libs.log.debug("Creating Stage1 task for : %d" %
                              index)

            tasks.append(GATKS1Application(
                vcf_group,
                index,
                **extra_args))

        return ParallelTaskCollection(tasks)


    def stage2(self):
        """
        Start this stage IIF stage1 all completed (i.e. no failures)
        * Use same GATK and goat.genome vesion as in stage0
        Combine 10 g.vcf files:
        take 10 + 3 and run GenotypeGVCF
        - add 3 more .g.vcf aggregated files (pased from CLI as option)
        output: 1 .cvf file (in result folder)
        Memory and walltime to be checked
        Example script:
java -jar /home/dleigh/GenomeAnalysisTK-3.1-1/GenomeAnalysisTK-3.4-46/GenomeAnalysisTK.jar \ # specifying the jar script
   -T GenotypeGVCFs \ # command within GATK
   -R /home/dleigh/goatgenome/01.GENOME/scaffold/goat_scaffoldFG_V1.1.normalised.22.07.fa \ # reference genome location
   -stand_call_conf 30 \ # constant
   -stand_emit_conf 10 \ # constant
   --variant /home/dleigh/demultiplexed.reads/GATK/combined.g.vcf \ # the input files
   --variant /home/dleigh/demultiplexed.reads/GATK/combined2.g.vcf \
   --variant /home/dleigh/demultiplexed.reads/GATK/combined3.g.vcf \
   -o /home/dleigh/demultiplexed.reads/GATK/genotyped.gvcf.vcf
        """
        vcfs_list = [ os.path.join(self.extra['S1_output'],vcfs) \
                      for vcfs in os.listdir(self.extra['S1_output'])]
        if 'S2_extra_input' in self.extra and self.extra['S2_extra_input']:
            vcfs_list.extend([ os.path.join(self.extra['S2_extra_input'],vcfs) \
                               for vcfs in os.listdir(self.extra['S2_extra_input']) ])

        extra_args = self.extra.copy()
        extra_args['jobname'] = "gatk-s2"

        extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                    extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                    extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                    extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                    extra_args['jobname'])

        gc3libs.log.debug("Creating Stage2 task")

        return GATKS2Application(
            vcfs_list,
            **extra_args)

class GgatkScript(SessionBasedScript):
    """
    For each param file (with '.mat' extension) found in the 'param folder',
    GscrScript extracts the corresponding index (from filename) and searches for
    the associated file in 'data folder'. For each pair ('param_file','data_file'),
    GscrScript generates execution Tasks.

    The ``gscr`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gscr`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            # application = GATKStagedTaskCollection,
            application = Application,
            stats_only_for = Application
            )

    def setup_args(self):

        self.add_param('bam_folder', type=str,
                       help="Location of samples files in .bam format.")

        self.add_param('gatk_sequence', type=str,
                       help="Comma separated list of GATK steps. "
                       "Valid values: %s" % gatk_steps_list)

    def setup_options(self):
        """
        * gatk
        * goat.genome
        """
        self.add_param("-g", "--gatk", metavar="PATH",
                       dest="gatk", default=None,
                       help="Location of alternative GATK jar file.")

        self.add_param("-G", "--goat", metavar="PATH",
                       dest="goat", default=None,
                       help="Location of alternative goat genome folder.")

        self.add_param("-S0O", "--s0-output",
                       metavar="PATH",
                       dest="S0_output",
                       default=os.path.join(os.environ['PWD'],"S0.out"),
                       help="Location were all outputs from Stage0"
                       " will be placed. Default: %(default)s.")

        self.add_param(
            "-S0M", "--s0-memory-per-task", dest="S0_memory",
            type=Memory, default=2 * GB,  # 2 GB
            metavar="GIGABYTES",
            help="Set the amount of memory required per S0 task;"
            " default: %(default)s. Specify this as an integral number"
            " followed by a unit, e.g., '512MB' or '4GB'.")

        self.add_param("-S1O", "--s1-output", metavar="PATH",
                       dest="S1_output",
                       default=os.path.join(os.environ['PWD'],"S1.out"),
                       help="Location were all outputs from Stage1"
                       " will be placed. Default: %(default)s.")

        self.add_param(
            "-S1M", "--s1-memory-per-task", dest="S1_memory",
            type=Memory, default=2 * GB,  # 2 GB
            metavar="GIGABYTES",
            help="Set the amount of memory required per S1 task;"
            " default: %(default)s. Specify this as an integral number"
            " followed by a unit, e.g., '512MB' or '4GB'.")

        self.add_param("-S1G", "--s1-group-factor", metavar="NUM",
                       dest="S1_group", default=30,
                       help="Group size for S1 aggregate task.")

        self.add_param("-S2O", "--s2-output", metavar="PATH",
                       dest="S2_output",
                       default=os.path.join(os.environ['PWD'],"S2.out"),
                       help="Location were all outputs from Stage2"
                       " will be placed. Default: %(default)s.")
        self.add_param(
            "-S2M", "--s2-memory-per-task", dest="S2_memory",
            type=Memory, default=2 * GB,  # 2 GB
            metavar="GIGABYTES",
            help="Set the amount of memory required per S2 task;"
            " default: %(default)s. Specify this as an integral number"
            " followed by a unit, e.g., '512MB' or '4GB'.")

        self.add_param("-S2E", "--s2-extra", metavar="PATH",
                       dest="S2_extra_input", default=None,
                       help="Location of the folder with Stage2 extra input files.")

    def parse_args(self):
        try:
            assert os.path.isdir(self.params.bam_folder), \
                "Input BAM folder %s not found." % self.params.bam_folder
            self.params.bam_folder = os.path.abspath(self.params.bam_folder)

            # make all path options ABS path
            self.params.S0_output = os.path.abspath(self.params.S0_output)
            self.params.S1_output = os.path.abspath(self.params.S1_output)
            self.params.S2_output = os.path.abspath(self.params.S2_output)

            # Check GATK sequence option
            self.params.gatk_sequence = [gatk_seq for gatk_seq in \
                                  self.params.gatk_sequence.split(',')
                                  if gatk_seq in gatk_steps_list]
            assert len(self.params.gatk_sequence) > 0,"Invalid GATK sequence: [%s]. " \
                "Please check input GATK sequence. " \
                "Valid values: %s" % (self.params.gatk_sequence,
                                      gatk_steps_list)
            self.params.gatk_sequence.sort()

            if self.params.S2_extra_input:
                assert os.path.isdir(self.params.S2_extra_input)
                self.params.S2_extra_input = os.path.abspath(self.params.S2_extra_input)

            assert int(self.params.S1_group), \
                "Invalid group size for S1 aggregate task: '%s'." % self.params.S1_group

            if self.params.goat:
                assert os.path.isdir(self.params.goat), \
                    "Location of alternative goat genome folder '%s' not found." \
                    % self.params.goat
                self.params.goat = os.path.abspat(self.params.goat)

        except AssertionError as ex:
            raise ValueError(ex.message)

    def new_tasks(self, extra):
        """
        For each of the network data and for each of the selected benchmarks,
        create a GscrApplication.

        First loop the input files, then loop the selected benchmarks
        """
        extra_args = extra.copy()
        extra_args.update(self.params.__dict__)
        # extra_args['S0_output_dir'] = os.path.abspath(self.params.S0_output)
        # extra_args['S1_output_dir'] = os.path.abspath(self.params.S1_output)
        # extra_args['S2_output_dir'] = os.path.abspath(self.params.S2_output)
        # extra_args['gatk_sequence'] = self.gatk_sequence
        # if self.params.S2_extra_input:
        #     extra_args['S2_extra_input'] = os.path.abspath(self.params.S2_extra_input)
        # extra_args['S1_group'] = int(self.params.S1_group)
        # if self.params.gatk:
        #     extra_args['gatk'] = self.params.gatk

        # if self.params.goat:
        #     extra_args['goat'] = self.params.goat

        # extra_args['output_dir'] = self.params.output

        # return [GATKStagedTaskCollection(self.params.bam_folder,
        #                                  **extra_args)]

        return [GATKSequentialTaskCollection(self.params.bam_folder,
                                             **extra_args)]



#! /usr/bin/env python

"""
Authentication support for accessing resources through the SSH protocol.
"""

# Copyright (C) 2009-2011, 2015, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
__docformat__ = 'reStructuredText'


import gc3libs
import gc3libs.defaults
from gc3libs.authentication import Auth
import gc3libs.exceptions


class SshAuth(object):

    def __init__(self,
                 # inherited from the generic `Auth` class
                 type,
                 # SSH-specific arguments
                 username,
                 keyfile=None,
                 port=None,
                 ssh_config=None,
                 timeout=None,
                 # extra arguments, if any
                 **extra):

        assert type == 'ssh'

        # this is required
        self.username = username

        # this can be set to None if no value is provided
        # (no sensible default)
        self.keyfile = keyfile

        if ssh_config is not None:
            self.ssh_config = ssh_config
        else:
            self.ssh_config = gc3libs.defaults.SSH_CONFIG_FILE

        # these need type conversion; if no value is supplied, use
        # `None` as doing otherwise would override settings from the
        # SSH config file in the `SshTransport` constructor.
        try:
            self.port = None if port is None else int(port)
        except (ValueError, TypeError) as err:
            raise gc3libs.exceptions.ConfigurationError(
                "Invalid `port` setting in SSH auth section.")
        try:
            self.timeout = None if timeout is None else float(timeout)
        except (ValueError, TypeError) as err:
            raise gc3libs.exceptions.ConfigurationError(
                "Invalid `timeout` setting in SSH auth section.")

        # everything else is just stored as-is
        self.__dict__.update(extra)

    def check(self):
        return True

    def enable(self):
        return True


Auth.register('ssh', SshAuth)


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="ssh",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gsisp.py -- Front-end script for running sisp
#   function with a given combination of input parameters.
#
#   Copyright (C) 2015, 2016  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gsisp.py --help`` for program usage
instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2016-01-15:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'

# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gsisp
    gsisp.Gsispingsispipt().run()

import os
import sys
import time
import tempfile
import re

import shutil
import random
import posix

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

PARAMETERS_FILE = "parameters.in"

## custom application class
class GsispingApplication(Application):
    """
    Custom class to wrap the execution of the sisp script.
    """
    application_name = 'gsisp'

    def __init__(self, input_folder, param_file, **extra_args):

        self.output_dir = extra_args['results_dir']

        inputs = dict()
        outputs = dict()
        executables = []
        self.jobname = extra_args['jobname']

        # Check if binary to be executed is provided as part of input arguments
        if 'sisp' in extra_args:
            inputs[os.path.abspath(extra_args["sisp"])] = "./sisp"
            executables.append("./sisp")

        arguments = "./sisp"
        inputs[param_file] = PARAMETERS_FILE

        # Set output
        outputs['output/'] = os.path.join(self.output_dir,"./output")

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gsisp.log',
            join=True,
            executables = executables,
            **extra_args)

    def termianted(self):
        """
        Check output folder
        """
        try:
            assert os.path.isdir(os.path.join(self.output_dir,"./output"))
        except AssertionError as ex:
            self.execution.returncode = (0, posix.EX_OSFILE)
            gc3libs.log.error("Job %s failed: %s" (self.jobname,ex.message))

class Gsispingsispipt(SessionBasedScript):
    """
    For each param file (with '.in' extension) found in the 'param folder',
    GsispScript generates execution Tasks.

    The ``gsisp`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gsisp`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GsispingApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GsispingApplication,
            )

    def setup_args(self):

        self.add_param('parameters', type=str,
                       help="Root localtion of input parameters. "
                       "Note: only folders containing `%s` files will be "
                       " considered." % PARAMETERS_FILE)

    def setup_options(self):
        self.add_param("-S", "--sisp", metavar="PATH",
                       dest="sisp", default=None,
                       help="Location of the sisp binary file.")

    def parse_args(self):
        """
        Check validity of input parameters and selected benchmark.
        """

        if not os.path.isdir(self.params.parameters):
            raise OSError("No such file or directory: %s ",
                          os.path.abspath(self.params.parameters))

        if self.params.sisp:
            if not os.path.isfile(self.params.sisp):
                raise gc3libs.exceptions.InvalidUsage("Sisp binary "
                                                      " file %s not found"
                                                      % self.params.sisp)
            else:
                self.params.sisp = os.path.abspath(self.params.sisp)

        # Walk through the input parameters folder and record all folders with `parameters.in` file

        try:
            self.folders = dict()
            for root,dir,files in os.walk(self.params.parameters):
                if PARAMETERS_FILE in files:
                    self.folders[os.path.abspath(root)] = os.path.join(root,files[files.index(PARAMETERS_FILE)])
            assert len(self.folders) > 0, "No valid input paramenters found"
        except AssertionError as ex:
            gc3libs.log.error(ex.message)

    def new_tasks(self, extra):
        """
        For each of the network data and for each of the selected benchmarks,
        create a GsispApplication.

        First loop the input files, then loop the selected benchmarks
        """
        tasks = []

        for parameter_folder in self.folders.keys():

            # Extract foldername
            jobname = os.path.basename(parameter_folder)

            extra_args = extra.copy()

            if self.params.sisp:
                extra_args['sisp'] = self.params.sisp

            extra_args['jobname'] = jobname
            extra_args['results_dir'] = parameter_folder
            tasks.append(GsispingApplication(
                parameter_folder,
                self.folders[parameter_folder],
                **extra_args))

        return tasks



#! /usr/bin/env python
#
#   gmhc_coev.py -- Front-end script for submitting multiple `MHC_coev` jobs to SMSCG.
#
#   Copyright (C) 2011-2012  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `MHC_coev` jobs to SMSCG.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gmhc_coev --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2011-07-22:
    * Use the ``APPS/BIO/MHC_COEV-040711`` run time tag to select execution sites.
  2011-06-15:
    * Initial release, forked off the ``ggamess`` sources.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == '__main__':
    import gmhc_coev
    gmhc_coev.GMhcCoevScript().run()


# std module imports
import csv
import glob
import math
import os
import re
import shutil
import sys
import time

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript
from collections import defaultdict
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import SequentialTaskCollection


## auxilirary functions

def _increase(value, inflection, ratio=2, increment=2):
    """
    Return increased `value`.

    The new value is `value` multiplied by `ratio` if the product is
    less than `inflection`; otherwise, it's `value` plus `increment`.
    In other words, values grow exponentially until the inflection
    point, after which they grow linearly.
    """
    if value > inflection:
        return (ratio * value)
    else:
        return (value + increment)


## custom application class

class GMhcCoevApplication(Application):
    """
    Custom class to wrap the execution of a single step of the
    ``MHC_coev_*`` program by A. B. Wilson and collaborators.
    """

    application_name = 'mhc_coev'

    def __init__(self,
                 N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last,
                 output_dir, latest_work=None, executable=None, **extra_args):
        extra_args.setdefault('requested_memory', 1*GB)
        extra_args.setdefault('requested_cores', 1)
        extra_args.setdefault('requested_architecture', Run.Arch.X86_64)
        # command-line parameters to pass to the MHC_coev_* program
        self.N = N
        self.p_mut_coeff = p_mut_coeff
        self.choose_or_rand = choose_or_rand
        self.sick_or_not = sick_or_not
        self.off_v_last = off_v_last
        if executable is not None:
            # use the specified executable
            executable_name = './' + os.path.basename(executable)
            inputs = { executable:os.path.basename(executable) }
        else:
            # use the default one provided by the RTE
            executable_name = '/$MHC_COEV'
            inputs = { }
        if latest_work is not None:
            inputs[latest_work] = 'latest_work.mat'
        Application.__init__(self, [
                                 executable_name,
                                 # use `single_run_time` as the
                                 # maximum allowed time before saving
                                 # the MatLab workspace, but allow 5
                                 # minutes for I/O before the job is
                                 # killed forcibly by the batch system
                                 extra_args.get('requested_walltime').amount(minutes) - 5,
                                 # rest of `MHC_coev` args as passed
                                 N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last,
                             ],
                             inputs = inputs,
                             outputs = gc3libs.ANY_OUTPUT,
                             output_dir = output_dir,
                             stdout = 'matlab.log',
                             stderr = 'matlab.err',
                             tags = [ 'TEST/MHC_COEV-040711ML2012' ],
                             **extra_args)


class GMhcCoevTask(SequentialTaskCollection):
    """
    Custom class to wrap the execution of the ``MHC_coev_*` program
    by A. B. Wilson and collaborators.

    Execution continues in steps of predefined duration until a
    specified number of generations have been computed and saved.  All
    the output files ever produced are collected in the path specified
    by the `output_dir` parameter to `__init__`.

    """

    def __init__(self, single_run_duration, generations_to_do,
                 N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last,
                 output_dir, executable=None, **extra_args):

        """Create a new task running an ``MHC_coev`` binary.

        Each binary is expected to run for `single_run_duration`
        minutes and dump its state in file ``latest_work.mat`` if it's
        not finished. This task will continue re-submitting the same
        executable together with the saved workspace until
        `generations_to_do` generations have been computed.

        :param single_run_duration:
          Duration of a single step (as a `gc3libs.quantity.Duration`:class: value)

        :param int generations_to_do:
          Count of generations that ``MHC_coev`` should simulate.

        :param              N: Passed unchanged to the MHC_coev program.
        :param    p_mut_coeff: Passed unchanged to the MHC_coev program.
        :param choose_or_rand: Passed unchanged to the MHC_coev program.
        :param    sick_or_not: Passed unchanged to the MHC_coev program.
        :param     off_v_last: Passed unchanged to the MHC_coev program.

        :param str output_dir:
          Path to a directory where output files
          from all runs should be collected.

        :param str executable:
          Path to the ``MHC_coev`` executable binary, or `None`
          (default) to specify that the default version available on
          the execution site should be used.

        """
        # remember values for later use
        self.executable = executable
        self.output_dir = output_dir
        # allow 5 extra minutes for final saving the MatLab workspace
        self.single_run_duration = single_run_duration + 5*minutes
        self.generations_to_do = generations_to_do
        self.p_mut_coeff = p_mut_coeff
        self.N = N
        self.choose_or_rand = choose_or_rand
        self.sick_or_not = sick_or_not
        self.off_v_last = off_v_last
        self.extra = extra_args

        self.generations_done = 0

        # make up a sensibel job name if there isn't one
        if 'jobname' in extra_args:
            self.jobname = extra_args['jobname']
        else:
            if self.executable is None:
                self.jobname = ('MHC_coev.%s.%s.%s.%s.%s'
                                % (N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last))
            else:
                os.path.basename(self.executable)

        # remove potentially conflicting kyword arguments
        extra_args.pop('output_dir', None)
        extra_args.pop('requested_walltime', None)

        # create initial task and register it
        initial_task = GMhcCoevApplication(
            N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last,
            output_dir = os.path.join(output_dir, 'tmp'),
            executable = self.executable,
            requested_walltime = self.single_run_duration,
            **extra_args)
        SequentialTaskCollection.__init__(self, [initial_task])


    # regular expression for extracting the generation no. from an output file name
    GENERATIONS_FILENAME_RE = re.compile(r'(?P<generation_no>[0-9]+)gen\.mat$')

    def next(self, done):
        """
        Analyze the retrieved output and decide whether to submit
        another run or not, depending on whether there is a
        ``latest_work.mat`` file.
        """
        task_output_dir = self.tasks[done].output_dir
        if not os.path.exists(task_output_dir):
            # no output, resubmit immediately
            return done
        stdout_filename = self.tasks[done].stdout
        stderr_filename = self.tasks[done].stderr
        last_run = self.tasks[done].execution
        exclude = [
            os.path.basename(self.tasks[done].arguments[0]),
            stdout_filename,
            stderr_filename,
            ]
        # move files one level up, except the ones listed in
        # `exclude`, and read the contents of STDOUT/STDERR into
        # `job_out` and `job_err`
        stdout_contents = ''
        stderr_contents = ''
        generation_files_count = 0
        for entry in os.listdir(task_output_dir):
            src_entry = os.path.join(task_output_dir, entry)
            # concatenate all output files together
            if entry == stdout_filename:
                gc3libs.utils.cat(src_entry, output=os.path.join(self.output_dir, entry), append=True)
                stdout_contents = gc3libs.utils.read_contents(src_entry)
            if entry == stderr_filename:
                gc3libs.utils.cat(src_entry, output=os.path.join(self.output_dir, entry), append=True)
                stderr_contents = gc3libs.utils.read_contents(src_entry)
            if entry in exclude or (entry.startswith('script.') and entry.endswith('.sh')):
                # delete entry and continue with next one
                os.remove(src_entry)
                continue
            # if `entry` is a generation output file, get the
            # generation no. and update the generation count
            match = GMhcCoevTask.GENERATIONS_FILENAME_RE.search(entry)
            if match:
                generation_files_count += 1
                generation_no = int(match.group('generation_no'))
                self.generations_done = max(self.generations_done, generation_no)
            # now really move file one level up
            dest_entry = os.path.join(self.output_dir, entry)
            if os.path.exists(dest_entry):
                # backup with numerical suffix
                gc3libs.utils.backup(dest_entry)
            os.rename(os.path.join(task_output_dir, entry), dest_entry)
        os.removedirs(task_output_dir)

        # scan for common failures and take correction measures
        report_filename = os.path.join(self.output_dir, stdout_filename)
        if os.path.exists(report_filename):
            report = open(report_filename, 'a')
        else:
            report = open(report_filename, 'w')
        report.write(
            "gmhc_coev: About last job: duration=%s, exitcode=%d, max_used_memory=%s\n"
            % (last_run.duration, last_run.exitcode, last_run.max_used_memory))


        # 0. did the job run at all?
        if last_run.duration < 5*minutes:
            report.write(
                "gmhc_coev: Job %s did not run at all,"
                " will try resubmitting to a different resource.\n"
                % (self,))
            # resubmit
            return done

        # 1. out of memory?
        elif (# generic memory error
            (last_run.max_used_memory >= self.extra['requested_memory'])
            # MATLAB-specific error condition
            or (last_run.exitcode == 255
                and ('Out of memory.' in stderr_contents
                     or 'MATLAB:nomem' in stderr_contents))):
            self.extra['requested_memory'] = _increase(self.extra['requested_memory'], 8*GB, increment=2*GB) # start growing linearly at 8GB
            report.write(
                "gmhc_coev: Possible out-of-memory condition detected in job %s,"
                " will request %s for next run.\n"
                % (self.tasks[done], self.requested_memory,))

        # 2. not enough time to compute even one generation?
        elif (generation_files_count == 0
              # be sure that the job has run for at least the expected time, +/- 30 minutes
              and (last_run.duration > (self.single_run_duration - 30*minutes))):
            self.single_run_duration = _increase(self.single_run_duration, 24*hours,
                                                 increase=4*hours)
            report.write(
                "gmhc_coev: Job %s could not complete 1 generation in the allotted time,"
                " will request %s for next run.\n"
                % (self.tasks[done], self.single_run_duration,))

        # if a `latest_work.mat` file exists, then we need
        # more time to compute the required number of generations
        latest_work = os.path.join(self.output_dir, 'latest_work.mat')
        if not os.path.exists(latest_work):
            latest_work = None
            report.write(
                "gmhc_coev: No `latest_work.mat` file was produced,"
                " next job will start from k=1.\n")
        if self.generations_done < self.generations_to_do:
            gc3libs.log.debug("Computed %s generations, %s to do; submitting another job."
                              % (self.generations_done, self.generations_to_do))
            report.write("gmhc_coev:"
                         " Computed %s generations, %s to do; submitting another job.\n"
                         % (self.generations_done, self.generations_to_do))
            report.close()
            self.add(
                GMhcCoevApplication(self.N, self.p_mut_coeff, self.choose_or_rand, self.sick_or_not, self.off_v_last,
                                    output_dir = os.path.join(self.output_dir, 'tmp'),
                                    latest_work = latest_work,
                                    executable = self.executable,
                                    requested_walltime = self.single_run_duration,
                                    **self.extra))
            return Run.State.RUNNING
        else:
            gc3libs.log.debug("Computed %s generations, no more to do."
                              % (self.generations_done))
            report.write("gmhc_coev:"
                         " Computed %s generations, no more to do.\n"
                         % (self.generations_done,))
            report.close()
            self.execution.returncode = self.tasks[done].execution.returncode
            return Run.State.TERMINATED


## main script class

class GMhcCoevScript(SessionBasedScript):
    """
Scan the specified INPUT directories recursively for executable files
whose name starts with the string ``MHC_coev``, and submit a job for
each file found; job progress is monitored and, when a job is done,
its ``.sav`` output files are retrieved back into the same directory
where the executable file is (this can be overridden with the ``-o``
option).

The ``gmhc_coev`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``gmhc_coev`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            input_filename_pattern = 'MHC_coev_*',
            application = GMhcCoevTask,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GMhcCoevTask,
            )

    def setup_options(self):
        self.add_param("-G", "--generations", metavar="NUM",
                       dest="generations", type=int, default=3000,
                       help="Compute NUM generations (default: 3000).")
        self.add_param("-x", "--executable", metavar="PATH",
                       dest="executable", default=None,
                       help="Path to the MHC_coev_* executable file.")


    def make_directory_path(self, pathspec, jobname):
        # XXX: Work around SessionBasedScript.process_args() that
        # apppends the string ``NAME`` to the directory path.
        # This is really ugly, but the whole `output_dir` thing needs to
        # be re-thought from the beginning...
        if pathspec.endswith('/NAME'):
            return pathspec[:-len('/NAME')]
        else:
            return pathspec


    def new_tasks(self, extra):
        # how many iterations are we already computing (per parameter set)?
        iters = defaultdict(lambda: 0)
        for task in self.session:
            name, instance = task.jobname.split('#')
            iters[name] = max(iters[name], int(instance))

        for path in self.params.args:
            if path.endswith('.csv'):
                try:
                    inputfile = open(path, 'r')
                except (OSError, IOError), ex:
                    self.log.warning("Cannot open input file '%s': %s: %s",
                                     path, ex.__class__.__name__, str(ex))
                try:
                    # the `csv.sniff()` function is confused by blank and comment lines,
                    # so we need to filter the input to build a correct sample
                    sample_lines = [ ]
                    while len(sample_lines) < 5:
                        line = inputfile.readline()
                        # exit at end of file
                        if line == '':
                            break
                        # ignore comment lines as they confuse `csv.sniff`
                        if line.startswith('#') or line.strip() == '':
                            continue
                        sample_lines.append(line)
                    csv_dialect = csv.Sniffer().sniff(str.join('', sample_lines))
                    self.log.debug("Detected CSV delimiter '%s'", csv_dialect.delimiter)
                except csv.Error:
                    # in case of any auto-detection failure, fall back to the default
                    self.log.warning("Could not determine field delimiter in file '%s',"
                                     " assuming it's a comma character (',').",
                                     path)
                    csv_dialect = 'excel'
                inputfile.seek(0)
                for lineno, row in enumerate(csv.reader(inputfile, csv_dialect)):
                    # ignore blank and comment lines (those that start with '#')
                    if len(row) == 0 or row[0].startswith('#'):
                        continue
                    try:
                        (iterno, N_str, p_mut_coeff_str, choose_or_rand_str, sick_or_not_str, off_v_last_str) = row
                    except ValueError:
                        self.log.error("Wrong format in line %d of file '%s':"
                                       " need 6 comma-separated values, but only got %d ('%s')."
                                       " Ignoring input line, fix it and re-run.",
                                       lineno+1, path, len(row), str.join(',', (str(x) for x in row)))
                        continue # with next `row`
                    # extract parameter values
                    try:
                        iterno = int(iterno)
                        N = GMhcCoevScript._parse_N(N_str)
                        p_mut_coeff = GMhcCoevScript._parse_p_mut_coeff(p_mut_coeff_str)
                        choose_or_rand = GMhcCoevScript._parse_choose_or_rand(choose_or_rand_str)
                        sick_or_not = GMhcCoevScript._parse_sick_or_not(sick_or_not_str)
                        off_v_last = GMhcCoevScript._parse_off_v_last(off_v_last_str)
                    except ValueError, ex:
                        self.log.warning("Ignoring line '%s' in input file '%s': %s",
                                         str.join(',', row), path, str(ex))
                        continue
                    basename = ('MHC_coev_' +
                                GMhcCoevScript._params_to_str(
                                    N, p_mut_coeff, choose_or_rand,
                                    sick_or_not, off_v_last))

                    # prepare job(s) to submit
                    if (iterno > iters[basename]):
                        self.log.info(
                                "Requested %d iterations for %s: %d already in session, preparing %d more",
                                iterno, basename, iters[basename], iterno - iters[basename])
                        for iter in range(iters[basename]+1, iterno+1):
                            kwargs = extra.copy()
                            base_output_dir = kwargs.pop('output_dir', self.params.output)
                            yield GMhcCoevTask(
                                self.params.walltime,    # single_run_duration
                                self.params.generations,
                                N,
                                p_mut_coeff,
                                choose_or_rand,
                                sick_or_not,
                                off_v_last,
                                executable=self.params.executable,
                                jobname=('%s#%d' % (basename, iter)),
                                output_dir = os.path.join(
                                    base_output_dir,
                                    ("N%d" % N),    # population size
                                    basename,       # params,
                                    str(iter),      # this iteration nr.
                                    ),
                                **kwargs)

            else:
                self.log.error("Ignoring input file '%s': not a CSV file.", path)


    ##
    ## INTERNAL METHODS
    ##


    _P_MUT_COEFF_RE = re.compile(r'((?P<num>[0-9]+)x)?10min(?P<exp>[0-9]+)')

    @staticmethod
    def _parse_p_mut_coeff(p_mut_coeff_str):
        try:
            return float(p_mut_coeff_str)
        except ValueError:
            match = GMhcCoevScript._P_MUT_COEFF_RE.match(p_mut_coeff_str)
            if not match:
                raise ValueError("Cannot parse P_MUT_COEFF expression '%s'" % p_mut_coeff_str)
            return float(match.group('num')) * 10.0**(-int(match.group('exp')))

    _N_RE = re.compile(r'N(?P<N>[0-9]+)')

    @staticmethod
    def _parse_N(N_str):
        try:
            return int(N_str)
        except ValueError:
            match = GMhcCoevScript._N_RE.match(N_str)
            if not match:
                raise ValueError("Cannot parse N expression '%s'" % N_str)
            return int(match.group('N'))

    @staticmethod
    def _parse_choose_or_rand(choose_or_rand_str):
        if choose_or_rand_str == "RM":
            return 1
        elif choose_or_rand_str == "DMAM":
            return 2
        elif choose_or_rand_str == "DMSSGD":
            return 3
        else:
            raise ValueError("Cannot parse CHOOSE_OR_RAND expression '%s'" % choose_or_rand_str)

    @staticmethod
    def _parse_sick_or_not(sick_or_not):
        if sick_or_not == "pat_on":
            return 1
        elif sick_or_not == "pat_off":
            return 0
        else:
            raise ValueError("Cannot parse SICK_OR_NOT expression '%s'" % sick_or_not)

    @staticmethod
    def _parse_off_v_last(off_v_last):
        try:
            return float(off_v_last)
        except ValueError:
            if not off_v_last.startswith("offval_"):
                raise ValueError("Cannot parse OFF_V_LAST expression '%s'" % off_v_last)
            off_v_last = off_v_last[7:]
            if off_v_last.startswith("0"):
                return float("0." + off_v_last[1:])
            elif off_v_last.startswith("1"):
                return 1.0
            else:
                raise ValueError("Cannot parse OFF_V_LAST expression '%s'" % off_v_last)

    @staticmethod
    def _string_to_params(s):
        """
        Return a tuple (N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last)
        obtained by parsing the string `s`.
        """
        p_mut_coeff, N, choose_or_rand, sick_or_not, off_v_last = s.split('__')

        return (GMhcCoevScript._parse_N(N),
                GMhcCoevScript._parse_p_mut_coeff(p_mut_coeff),
                GMhcCoevScript._parse_choose_or_rand(choose_or_rand),
                GMhcCoevScript._parse_sick_or_not(sick_or_not),
                GMhcCoevScript._parse_off_v_last(off_v_last))

    @staticmethod
    def _N_to_str(N):
        return ("N%d" % N)

    @staticmethod
    def _p_mut_coeff_to_str(p_mut_coeff):
        """
        Print the floating-point number `p_mut_coeff` as ``1x10min3``
        instead of the usual ``1e-3``.

        Examples::

          >>> _p_mut_coeff_to_str(0.005)
          '5x10min3'
          >>> _p_mut_coeff_to_str(10)
          '1x10plus1'
          >>> _p_mut_coeff_to_str(0.1234)
          '1.234x10min1'

        """
        exponent = math.log10(p_mut_coeff)
        mantissa = p_mut_coeff * (10 ** math.ceil(-exponent))
        # format output string
        if (mantissa == int(mantissa)):
            # no fractional part
            result = str(int(mantissa))
        else:
            result = str(mantissa)
        if exponent != 0:
            if exponent < 0:
                result += ('x10min%d' % (-int(exponent),))
            else:
                result += ('x10plus%d' % (int(exponent),))
        return result

    @staticmethod
    def _choose_or_rand_to_str(choose_or_rand):
        if choose_or_rand == 1:
            return "RM"
        elif choose_or_rand == 2:
            return "DMAM"
        elif choose_or_rand == 3:
            return "DMSSGD"
        else:
            raise ValueError("Valid values for `choose_or_rand` are: 1,2,3;"
                             " got '%s' instead" % choose_or_rand)

    @staticmethod
    def _sick_or_not_to_str(sick_or_not):
        if sick_or_not:
            return "pat_on"
        else:
            return "pat_off"

    @staticmethod
    def _off_v_last_to_str(off_v_last):
        if off_v_last < 0.0 or off_v_last > 1.0:
            raise ValueError("Parameter `off_v_last` must be in range 0.0 to 1.0;"
                             " got `%s`instead." % off_v_last)
        if off_v_last == 1.0:
            return "offval_1"
        else:
            # off_v_last == 0.xxxx

            # XXX: Python switches to scientific notation for
            # N<0.00001; the funny %-hack below ensures that the
            # dot+digits notation is used instead; starting Python
            # 2.6, the `format` function is available and should be
            # preferred:: off_v_last_str = format(off_v_last, 'f')
            off_v_last_str = ("%.16f" % off_v_last).rstrip('0')
            return ('offval_0' + off_v_last_str[2:])

    @staticmethod
    def _params_to_str(N, p_mut_coeff, choose_or_rand, sick_or_not, off_v_last):
        return str.join('__', [
            GMhcCoevScript._p_mut_coeff_to_str(p_mut_coeff),
            GMhcCoevScript._N_to_str(N),
            GMhcCoevScript._choose_or_rand_to_str(choose_or_rand),
            GMhcCoevScript._sick_or_not_to_str(sick_or_not),
            GMhcCoevScript._off_v_last_to_str(off_v_last),
            ])



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import range
from builtins import object
__docformat__ = 'reStructuredText'

import pytest

from gc3libs.persistence.idfactory import IdFactory


class DummyObject(object):
    pass


def test_new_item():
    idfactory = IdFactory()
    ids = []
    dummy = DummyObject()
    for i in range(10):
        ids.append(idfactory.new(dummy))
    assert len(ids) == len(set(ids))

    # reserve is tested only in order to check if we get an error calling
    # it...
    idfactory.reserve(5)

@pytest.mark.skip("Code currently bugged, cfr. issue #608")
def test_custom_next_id():
    class next_id(object):

        def __init__(self):
            self.curid = -1

        def __call__(self):
            self.curid += 1
            return self.curid

    idfactory = IdFactory(next_id_fn=next_id())

    ids = []
    dummy = DummyObject()
    for i in range(10):
        ids.append(idfactory.new(dummy))

    assert len(ids) == len(set(ids))

    for i in range(len(ids)):
        assert ids[i] == "DummyObject.%d" % i


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="test_idfactory",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



"""
A backport of the Python standard `collections` package, providing
`OrderedDict` also on Python 2.6.

This package actually imports your Python `collections`, and adds
its own version of `OrderedDict` only if missing.
"""

from __future__ import absolute_import, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from collections import *  # noqa

try:
    OrderedDict
except:
    # Backport of OrderedDict() class that runs on Python 2.4, 2.5,
    # 2.6, 2.7 and pypy.
    #
    # Passes Python2.7's test suite and incorporates all the latest
    # updates.

    try:
        from _thread import get_ident as _get_ident
    except ImportError:
        from _dummy_thread import get_ident as _get_ident

    try:
        from _abcoll import KeysView, ValuesView, ItemsView
    except ImportError:
        pass

    class OrderedDict(dict):

        'Dictionary that remembers insertion order'
        # An inherited dict maps keys to values.  The inherited dict
        # provides __getitem__, __len__, __contains__, and get.  The
        # remaining methods are order-aware.  Big-O running times for
        # all methods are the same as for regular dictionaries.

        # The internal self.__map dictionary maps keys to links in a
        # doubly linked list.  The circular doubly linked list starts
        # and ends with a sentinel element.  The sentinel element
        # never gets deleted (this simplifies the algorithm).  Each
        # link is stored as a list of length three: [PREV, NEXT, KEY].

        def __init__(self, *args, **kwds):
            '''Initialize an ordered dictionary.  Signature is the same as for
            regular dictionaries, but keyword arguments are not recommended
            because their insertion order is arbitrary.

            '''
            if len(args) > 1:
                raise TypeError(
                    'expected at most 1 arguments, got %d' %
                    len(args))
            try:
                self.__root
            except AttributeError:
                self.__root = root = []                     # sentinel node
                root[:] = [root, root, None]
                self.__map = {}
            self.__update(*args, **kwds)

        def __setitem__(self, key, value, dict_setitem=dict.__setitem__):
            'od.__setitem__(i, y) <==> od[i]=y'
            # Setting a new item creates a new link which goes at the
            # end of the linked list, and the inherited dictionary is
            # updated with the new key/value pair.
            if key not in self:
                root = self.__root
                last = root[0]
                last[1] = root[0] = self.__map[key] = [last, root, key]
            dict_setitem(self, key, value)

        def __delitem__(self, key, dict_delitem=dict.__delitem__):
            'od.__delitem__(y) <==> del od[y]'
            # Deleting an existing item uses self.__map to find the
            # link which is then removed by updating the links in the
            # predecessor and successor nodes.
            dict_delitem(self, key)
            link_prev, link_next, key = self.__map.pop(key)
            link_prev[1] = link_next
            link_next[0] = link_prev

        def __iter__(self):
            'od.__iter__() <==> iter(od)'
            root = self.__root
            curr = root[1]
            while curr is not root:
                yield curr[2]
                curr = curr[1]

        def __reversed__(self):
            'od.__reversed__() <==> reversed(od)'
            root = self.__root
            curr = root[0]
            while curr is not root:
                yield curr[2]
                curr = curr[0]

        def clear(self):
            'od.clear() -> None.  Remove all items from od.'
            try:
                for node in self.__map.values():
                    del node[:]
                root = self.__root
                root[:] = [root, root, None]
                self.__map.clear()
            except AttributeError:
                pass
            dict.clear(self)

        def popitem(self, last=True):
            '''od.popitem() -> (k, v), return and remove a (key, value) pair.
            Pairs are returned in LIFO order if last is true or FIFO
            order if false.

            '''
            if not self:
                raise KeyError('dictionary is empty')
            root = self.__root
            if last:
                link = root[0]
                link_prev = link[0]
                link_prev[1] = root
                root[0] = link_prev
            else:
                link = root[1]
                link_next = link[1]
                root[1] = link_next
                link_next[0] = root
            key = link[2]
            del self.__map[key]
            value = dict.pop(self, key)
            return key, value

        # -- the following methods do not depend on the internal structure --

        def keys(self):
            'od.keys() -> list of keys in od'
            return list(self)

        def values(self):
            'od.values() -> list of values in od'
            return [self[key] for key in self]

        def items(self):
            'od.items() -> list of (key, value) pairs in od'
            return [(key, self[key]) for key in self]

        def iterkeys(self):
            'od.iterkeys() -> an iterator over the keys in od'
            return iter(self)

        def itervalues(self):
            'od.itervalues -> an iterator over the values in od'
            for k in self:
                yield self[k]

        def iteritems(self):
            'od.iteritems -> an iterator over the (key, value) items in od'
            for k in self:
                yield (k, self[k])

        def update(*args, **kwds):
            '''od.update(E, **F) -> None.
            Update od from dict/iterable E and F.

            If E is a dict instance, does:

                for k in E: od[k] = E[k]

            If E has a .keys() method, does:

                for k in E.keys(): od[k] = E[k]

            Or if E is an iterable of items, does:

                for k, v in E: od[k] = v

            In either case, this is followed by:

                for k, v in F.items(): od[k] = v

            '''
            if len(args) > 2:
                raise TypeError('update() takes at most 2 positional '
                                'arguments (%d given)' % (len(args),))
            elif not args:
                raise TypeError('update() takes at least 1 argument (0 given)')
            self = args[0]
            # Make progressively weaker assumptions about "other"
            other = ()
            if len(args) == 2:
                other = args[1]
            if isinstance(other, dict):
                for key in other:
                    self[key] = other[key]
            elif hasattr(other, 'keys'):
                for key in list(other.keys()):
                    self[key] = other[key]
            else:
                for key, value in other:
                    self[key] = value
            for key, value in list(kwds.items()):
                self[key] = value

        # let subclasses override update without breaking __init__
        __update = update

        __marker = object()

        def pop(self, key, default=__marker):
            '''od.pop(k[,d]) -> v, remove specified key and return the
            corresponding value.  If key is not found, d is returned
            if given, otherwise KeyError is raised.

            '''
            if key in self:
                result = self[key]
                del self[key]
                return result
            if default is self.__marker:
                raise KeyError(key)
            return default

        def setdefault(self, key, default=None):
            """
            od.setdefault(k[,d]) -> od.get(k,d), also set od[k]=d if k not in
            od

            """
            if key in self:
                return self[key]
            self[key] = default
            return default

        def __repr__(self, _repr_running={}):
            'od.__repr__() <==> repr(od)'
            call_key = id(self), _get_ident()
            if call_key in _repr_running:
                return '...'
            _repr_running[call_key] = 1
            try:
                if not self:
                    return '%s()' % (self.__class__.__name__,)
                return '%s(%r)' % (self.__class__.__name__, list(self.items()))
            finally:
                del _repr_running[call_key]

        def __reduce__(self):
            'Return state information for pickling'
            items = [[k, self[k]] for k in self]
            inst_dict = vars(self).copy()
            for k in vars(OrderedDict()):
                inst_dict.pop(k, None)
            if inst_dict:
                return (self.__class__, (items,), inst_dict)
            return self.__class__, (items,)

        def copy(self):
            'od.copy() -> a shallow copy of od'
            return self.__class__(self)

        @classmethod
        def fromkeys(cls, iterable, value=None):
            '''
            OD.fromkeys(S[, v]) -> New ordered dictionary with keys from S and
            values equal to v (which defaults to None).

            '''
            d = cls()
            for key in iterable:
                d[key] = value
            return d

        def __eq__(self, other):
            '''od.__eq__(y) <==> od==y.  Comparison to another OD is
            order-sensitive while comparison to a regular mapping is
            order-insensitive.

            '''
            if isinstance(other, OrderedDict):
                return len(self) == len(
                    other) and list(self.items()) == list(other.items())
            return dict.__eq__(self, other)

        def __ne__(self, other):
            return not self == other

        # -- the following methods are only used in Python 2.7 --

        def viewkeys(self):
            "od.viewkeys() -> a set-like object providing a view on od's keys"
            return KeysView(self)

        def viewvalues(self):
            "od.viewvalues() -> an object providing a view on od's values"
            return ValuesView(self)

        def viewitems(self):
            """
            od.viewitems() -> a set-like object providing a view on od's items
            """
            return ItemsView(self)



#! /usr/bin/env python
#
"""
Unit tests for the EC2 backend.
"""
# Copyright (C) 2011-2013, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
from builtins import object
__docformat__ = 'reStructuredText'

# stdlib imports
import shutil
import tempfile

# 3rd party imports
import pytest

# The EC2 backend might not be installed (it's currently marked as
# optional in `setup.py`), so skip these tests altogether if there is any error
boto = pytest.importorskip("boto")

# local imports
from gc3libs.backends.ec2 import VMPool
import gc3libs.exceptions


class _MockVM(object):

    def __init__(self, id, **extra):
        self.id = id
        for k, v in extra.items():
            setattr(self, k, v)


class TestVMPool(object):

    # XXX: the `get*` methods are not tested here (yet) as they
    # require mocking the `boto` library!

    @pytest.fixture(autouse=True)
    def setup(self):
        # for persistence tests
        self.tmpdir = tempfile.mkdtemp()

        # an empty VMPool
        self.pool0 = VMPool(self.tmpdir + '/pool0', None)
        # VMpool with 1 VM
        self.pool1 = VMPool(self.tmpdir + '/pool1', None)
        self.vm1 = _MockVM('a')
        self.pool1.add_vm(self.vm1)
        # VMpool with 2 VM
        self.pool2 = VMPool(self.tmpdir + '/pool2', None)
        self.vm2 = _MockVM('b')
        self.pool2.add_vm(self.vm1)
        self.pool2.add_vm(self.vm2)

        yield

        shutil.rmtree(self.tmpdir)

    def test_empty_vmpool(self):
        assert self.pool0._vm_ids == set([])
        assert self.pool0._vm_cache == {}

    def test_vmpool_one_vm(self):
        assert self.pool1._vm_ids == set(['a'])
        assert self.pool1._vm_cache == {'a': self.vm1}

    def test_vmpool_two_vms(self):
        assert self.pool2._vm_ids == set(['a', 'b'])
        assert self.pool2._vm_cache == {'a': self.vm1, 'b': self.vm2}

    def test_repr(self):
        # representation of empty sets differs in Py2 and Py3 ...
        assert repr(self.pool0) in ["set([])", "set()"]
        # ... also representation of unicode strings differs on Py2 and Py3
        assert repr(self.pool1) in ["{'a'}", "set(['a'])", "set([u'a'])"]
        # ...and also sets do not have predictable representation
        assert repr(self.pool2) in [
            # Py2
            "set(['a', 'b'])",
            "set(['b', 'a'])",
            "set([u'a', u'b'])",
            "set([u'b', u'a'])",
            # Py3
            "{'a', 'b'}",
            "{'b', 'a'}",
        ]


    def test_str(self):
        # representation of empty sets differs in Py2 and Py3 ...
        assert str(self.pool0) in [
            "VMPool('pool0') : set()",
            "VMPool('pool0') : set([])",
        ]
        # representation of unicode strings differs on Py2 and Py3
        assert str(self.pool1) in [
            "VMPool('pool1') : {'a'}",
            "VMPool('pool1') : set(['a'])",
            "VMPool('pool1') : set([u'a'])",
        ]
        # also sets do not have predictable representation
        assert str(self.pool2) in [
            # Py2
            "VMPool('pool2') : set(['a', 'b'])",
            "VMPool('pool2') : set(['b', 'a'])",
            "VMPool('pool2') : set([u'a', u'b'])",
            "VMPool('pool2') : set([u'b', u'a'])",
            # Py3
            "VMPool('pool2') : {'a', 'b'}",
            "VMPool('pool2') : {'b', 'a'}",
        ]

    def test_add_remove(self):
        VM_ID = 'x'
        vm = _MockVM(VM_ID)
        for pool in self.pool0, self.pool1, self.pool2:
            L = len(pool)
            # test add
            pool.add_vm(vm)
            assert VM_ID in pool
            assert len(pool) == L + 1
            # test remove
            pool.remove_vm(VM_ID)
            assert not VM_ID in pool
            assert len(pool) == L

    def test_add_delete(self):
        VM_ID = 'x'
        vm = _MockVM(VM_ID)
        for pool in self.pool0, self.pool1, self.pool2:
            L = len(pool)
            # test add
            pool.add_vm(vm)
            assert VM_ID in pool
            assert len(pool) == L + 1
            # test delete
            del pool[VM_ID]
            assert not VM_ID in pool
            assert len(pool) == L

    def test_len(self):
        for l, pool in enumerate([self.pool0, self.pool1, self.pool2]):
            assert len(pool) == l

    def test_iter(self):
        """Check that `VMPool.__iter__` iterates over VM IDs."""
        for n, pool in enumerate([self.pool0, self.pool1, self.pool2]):
            assert list(iter(pool)) in (
                list('ab'[:n]),
                list('ba'[:n]),
            )

    def test_get_vm_in_cache(self):
        assert self.vm1 == self.pool1['a']

        assert self.vm1 == self.pool2['a']
        assert self.vm2 == self.pool2['b']

    def test_get_vm_not_in_cache_and_no_connection(self):
        # clone pool2 from disk copy
        pool = VMPool(self.pool2.path, None)
        with pytest.raises(gc3libs.exceptions.UnrecoverableError):
            # pylint: disable=pointless-statement
            pool['a']

    def test_get_all_vms(self):
        all_vms1 = self.pool1.get_all_vms()
        assert self.vm1 in all_vms1

        all_vms2 = self.pool2.get_all_vms()
        assert self.vm1 in all_vms2
        assert self.vm2 in all_vms2

    def test_lookup_is_get_vm(self):
        assert self.pool1.get_vm('a') == self.pool1['a']

        assert self.pool2.get_vm('a') == self.pool2['a']
        assert self.pool2.get_vm('b') == self.pool2['b']

    def test_reload(self):
        # simulate stopping program by deleting object and cloning a
        # copy from disk
        for pool in self.pool0, self.pool1, self.pool2:
            ids = pool._vm_ids
            path = pool.path
            del pool
            pool = VMPool(path, None)
            assert pool._vm_ids == ids
            assert pool._vm_ids is not ids

    def test_save_then_load_empty_vmpool(self):
        self.pool0.save()

        pool = VMPool(self.pool0.path, None)
        assert pool.name == self.pool0.name
        assert pool.conn == None
        assert pool._vm_cache == {}
        assert pool._vm_ids == self.pool0._vm_ids

    def test_save_then_load_nonempty_vmpool(self):
        self.pool2.save()

        pool = VMPool(self.pool2.path, None)
        assert pool.name == self.pool2.name
        # the list of VM ids should have been persisted ...
        assert pool._vm_ids == self.pool2._vm_ids
        # ... but the VM cache is now empty
        assert pool._vm_cache == {}
        # ...and so is the connection
        assert pool.conn == None

    @pytest.mark.skip("This test must be rewritten")
    def test_update(self):
        # clone pool2
        pool = VMPool(self.pool2.path, None)
        vm3 = _MockVM('c')
        pool.add_vm(vm3)
        self.pool2.update()
        assert 'c' in self.pool2
        # cannot test the following without a connection:
        # assert_equal(self.pool2['c'], vm3)

    def test_update_remove(self):
        # clone pool2
        pool = VMPool(self.pool2.path, None)
        pool.remove_vm('b')
        self.pool2.update(remove=True)
        assert 'b' not in self.pool2
        assert 'a' in self.pool2


if "__main__" == __name__:
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
#   gspg.py -- Front-end script for evaluating R-based 'weight'
#   function over a large dataset.
#
#   Copyright (C) 2011, 2012  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `R` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gspg.py --help`` for program usage
instructions.

Input parameters consists of:
:param str edges file: Path to an .csv file containing input data in
the for of:
    X1   X2
1  id1  id2
2  id1  id3
3  id1  id4

...

XXX: To be clarified:
. When input files should be removed ?
. What happen if an error happen at merging time ?
. Should be possible to re-run a subset of the initial chunk list
without re-creating a new session ?
e.g. adding a new argument accepting chunk ranges (-R 3000:7500)
This would trigger the re-run of the whole workflow only
for lines between 3000 and 7500
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2013-07-03:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@gc3.uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gspg
    gspg.GspgScript().run()

import os
import sys
import time
import tempfile

import shutil
import pandas
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

## custom application class
class GspgApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gspg'

    def __init__(self, input_file, **extra_args):


        executables = []
        inputs = dict()

        inputs[input_file] = "./input.csv"
        gspg_wrapper = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gspg_wrapper.py")

        inputs[gspg_wrapper] = "./wrapper.py"


        self.output_folder = "./results"

        # setup output references
        # self.result_dir = result_dir
        self.output_dir = extra_args['output_dir']
        self.output_filename = 'results-%s.csv' % extra_args['index_chunk']
        self.output_file = os.path.join(self.output_dir,self.output_filename)

        outputs = [("./results.csv",self.output_filename)]

        arguments ="python ./wrapper.py ./input.csv "

        # check the optional inputs

        if extra_args.has_key('ctx'):
            inputs[extra_args['ctx']] = "./bin/ctx.p4"
            arguments +=  "--ctx ./bin/ctx.p4 "
            executables.append("./bin/ctx.p4")

        gc3libs.log.debug("Creating application for executing: %s",
                          arguments)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gspg.log',
            join=True,
            executables = executables,
            **extra_args)

    def terminated(self):
        """
        Check whether output file has been properly created
        """
        pass

class GspgScript(SessionBasedScript):
    """
    Takes 1 .csv input file containing the list of parameters to be passed
    to the `ctx-linkdyn-ordprm-sirs.p4` application.
    Each line of the input .csv file correspond to the parameter list to be
    passed to a single `ctx-linkdyn-ordprm-sirs.p4` execution. For each line
    of the input .csv file a GspgApplication needs to be generated (depends on
    chunk value passed as part of the input options).
    Splits input .csv file into smaller chunks, each of them of size
    'self.params.chunk_size'.
    Then submits one execution for each of the created chunked files.

    The ``gspg`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gspg`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gspg``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GspgApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GspgApplication,
            )

    def setup_args(self):

        self.add_param('csv_input_file', type=str,
                       help="Input .csv file")

    def setup_options(self):

        self.add_param("-M", "--ctx", metavar="[PATH]",
                       dest="ctx", default=None,
                       help="Location of local copy of ctx binary")

        self.add_param("-k", "--chunk", metavar="INT", type=int,
                       dest="chunk_size", default=1000,
                       help="How to split the edges input data set.")

    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to command_file should also be valid.
        """

        try:
            assert os.path.isfile(self.params.csv_input_file)
        except ValueError:
            raise gc3libs.exceptions.InvalidUsage(
                "Input CSV file %s not found" % self.params.csv_input_file)

        if self.params.ctx:
            try:
                assert os.path.isfile(self.params.ctx)
            except ValueError:
                raise gc3libs.exceptions.InvalidUsage(
                    "CTX binary file %s not found" % self.params.ctx)

        # Verify that 'self.params.chunk_size' is int
        try:
            assert int(self.params.chunk_size)
        except ValueError:
            raise gc3libs.exceptions.InvalidUsage(
                "-k option accepts only numbers")

    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new GcgpsTask
        """
        tasks = []

        for (input_file, index_chunk) in self._generate_chunked_files_and_list(self.params.csv_input_file,
                                                                              self.params.chunk_size):

            jobname = "gspg-%s" % (str(index_chunk))

            extra_args = extra.copy()

            extra_args['index_chunk'] = str(index_chunk)
            extra_args['chunk_size'] = int(self.params.chunk_size)

            extra_args['jobname'] = jobname

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

            if self.params.ctx:
                extra_args['ctx'] = self.params.ctx

            self.log.debug("Creating Application for index : %d - %d" %
                           (index_chunk, (index_chunk + self.params.chunk_size)))

            tasks.append(GspgApplication(
                    input_file,
                    **extra_args))

        return tasks

    def _generate_chunked_files_and_list(self, file_to_chunk, chunk_size=2):
        """
        Takes a file_name as input and a defined chunk size
        ( uses 1000 as default )
        returns a list of filenames 1 for each chunk created and a corresponding
        index reference
        e.g. ('/tmp/chunk2800.csv,2800) for the chunk segment that goes
        from 2800 to (2800 + chunk_size)
        """

        chunks = []
        chunk_files_dir = os.path.join(self.session.path,"tmp")

        # creating 'chunk_files_dir'
        if not(os.path.isdir(chunk_files_dir)):
            try:
                os.mkdir(chunk_files_dir)
            except OSError, osx:
                gc3libs.log.error("Failed while creating tmp folder %s. " % chunk_files_dir +
                                  "Error %s." % str(osx) +
                                  "Using default '/tmp'")
                chunk_files_dir = "/tmp"

        # XXX: by convenction, 1st row contains headers
        reader = pandas.read_csv(file_to_chunk, header=0, chunksize=chunk_size)

        index = 0
        for chunk in reader:
            index += 1
            filename = "%s/chunk_%s.csv" % (chunk_files_dir,index)
            chunk.to_csv(filename, header=True, index=False)
            chunks.append((filename,index))

        return chunks



#! /usr/bin/env python
#
#   gpfi.py -- Front-end script for evaluating Matlab functions
#   function over a large number of parameters.
#
#   Copyright (C) 2015, 2016  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify it under the terms of the GNU General Public License as
#   published by the Free Software Foundation, either version 3 of
#   the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple Matlab jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gpfi.py --help`` for program usage instructions.

Input parameters consists of: model input_csv: Path to an .csv file
containing input parameters
Example:
0.1 0.9
0.2 0.8
...
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2016-04-20:
  * Initial version
"""
__author__ = 'Tyanko Aleksiev <tyanko.aleksiev@uzh.ch>'
__docformat__ = 'reStructuredText'


if __name__ == "__main__":
    import gpfi
    gpfi.GpfiScript().run()

import os
import sys
import time
import tempfile

import shutil

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

## custom application class
class GpfiApplication(Application):
    """
    Custom class to wrap the execution of the matlab script passed in src_dir.
    """
    application_name = 'gpfi'

    def __init__(self, parameter, model, **extra_args):

        inputs = dict()
        outputs = dict()

        # execution wrapper needs to be added anyway
        gpfi_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gpfi.sh")
        inputs[gpfi_wrapper_sh] = os.path.basename(gpfi_wrapper_sh)
        inputs[model] = os.path.basename(model)

        _command = "%s %s" % (os.path.basename(gpfi_wrapper_sh), ' '.join(str(x) for x in parameter))

        Application.__init__(
            self,
            arguments = _command,
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'gpfi.log',
            join=True,
            executables = "./%s" % os.path.basename(gpfi_wrapper_sh),
            **extra_args)


class GpfiScript(SessionBasedScript):
    """
    Parse the input .csv file. For each line in the input .csv file
    create a different Matlab execution.

    The ``gpfi`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gpfi`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gpfi``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GpfiApplication,
            stats_only_for = GpfiApplication,
            )

    def setup_args(self):
        self.add_param('model', type=str, help="Location of the matlab scripts and related MAtlab functions. Default: None")

    def setup_args(self):
        self.add_param('csv_input_file', type=str, help="Input .csv file")

    def parse_args(self):
        """
        Check presence of input folder (should contains matlab scripts).
        path to command_file should also be valid.
        """
        assert os.path.isfile(self.params.csv_input_file), \
        "Input CSV file %s not found" % self.params.csv_input_file

    def new_tasks(self, extra):
        """
        For each line of the input .csv file generate an execution Task
        """
        tasks = []

        for parameter in self._enumerate_csv(self.params.csv_input_file):
            parameter_str = '.'.join(str(x) for x in parameter)
            jobname = "gpfi-%s" % parameter_str

            extra_args = extra.copy()

            extra_args['jobname'] = jobname

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

            self.log.debug("Creating Application for parameter : %s" %
                           (parameter_str))

            tasks.append(GpfiApplication(
                    parameter,
                    self.param.model,
                    **extra_args))

        return tasks

    def _enumerate_csv(self, csv_input):
        """
        For each line of the input .csv file return list of parameters
        """
        csv_file = open(csv_input, 'rb')
        csv_reader = csv.reader(csv_file)
        next(csv_reader, None)
        for row in reader:
            yield row



#! /usr/bin/env python
#
"""
SQL-based storage of GC3pie objects.
"""
# Copyright (C) 2011-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import absolute_import, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import str
__docformat__ = 'reStructuredText'


# stdlib imports
from contextlib import closing
from io import BytesIO
import os
from urllib.parse import parse_qs
from warnings import warn
from weakref import WeakValueDictionary

import sqlalchemy as sqla
import sqlalchemy.sql as sql

# GC3Pie interface
from gc3libs import Run
import gc3libs.exceptions
from gc3libs.url import Url
import gc3libs.utils
from gc3libs.utils import same_docstring_as

from gc3libs.persistence.idfactory import IdFactory
from gc3libs.persistence.serialization import make_pickler, make_unpickler
from gc3libs.persistence.store import Store


# uncomment lines containing `_lvl` to show nested save/loads in logs
#_lvl = ''


class IntId(int):

    def __new__(cls, prefix, seqno):
        return int.__new__(cls, seqno)

    def __getnewargs__(self):
        return (None, int(self))


class SqlStore(Store):

    """
    Save and load objects in a SQL db, using python's `pickle` module
    to serialize objects into a specific field.

    Access to the DB is done via SQLAlchemy module, therefore any
    driver supported by SQLAlchemy will be supported by this class.

    The `url` argument is used to access the store. It is supposed to
    be a `gc3libs.url.Url`:class: class, and therefore may contain
    username, password, host and port if they are needed by the db
    used.

    The `table_name` argument is the name of the table to create. By
    default it's ``store``.  Alternatively, the table name can be
    given in the "fragment" part of the database URL, as
    ``#table=...`` (replace ``...`` with the actual table name).  The
    constructor argument takes precedence over the table name
    specified in the DB URL.

    The constructor will create the `table_name` table if it does not
    exist, but if there already is such a table it will assume that
    its schema is compatible with our needs. A minimal table schema
    is as follows::

        +-----------+--------------+------+-----+---------+
        | Field     | Type         | Null | Key | Default |
        +-----------+--------------+------+-----+---------+
        | id        | int(11)      | NO   | PRI | NULL    |
        | data      | blob         | YES  |     | NULL    |
        | state     | varchar(128) | YES  |     | NULL    |
        +-----------+--------------+------+-----+---------+

    The meaning of the fields is:

    - `id`: this is the id returned by the `save()` method and
      uniquely identifies a stored object.

    - `data`: serialized Python object.

    - `state`: if the object is a `Task`:class: instance, this will be
      its current execution state.

    The `extra_fields` constructor argument is used to extend the
    database. It must contain a mapping `*column*: *function*`
    where:

    - *column* is a `sqlalchemy.Column` object.

    - *function* is a function which takes the object to be saved as
      argument and returns the value to be stored into the
      database. Any exception raised by this function will be
      *ignored*.  Classes `GetAttribute`:class: and `GetItem`:class:
      in module `get`:mod: provide convenient helpers to save object
      attributes into table columns.

    For each extra column the `save()` method will call the
    corresponding *function* in order to get the correct value to
    store into the DB.

    Any extra keyword arguments are ignored for compatibility with
    `FilesystemStore`:class:.
    """

    def __init__(self, url, table_name=None, idfactory=None,
                 extra_fields=None, create=True, **extra_args):
        """
        Open a connection to the storage database identified by `url`.

        DB backend (MySQL, psql, sqlite3) is chosen based on the
        `url.scheme` value.
        """
        super(SqlStore, self).__init__(url)
        if self.url.fragment:
            kv = parse_qs(self.url.fragment)
        else:
            kv = {}

        # init static public args
        self.idfactory = idfactory or IdFactory(id_class=IntId)

        url_table_names = kv.get('table')
        if url_table_names:
            url_table_name = url_table_names[-1]  # last wins
        else:
            url_table_name = ''
        if table_name is None:
            self.table_name = url_table_name or "store"
        else:
            if table_name != url_table_name:
                gc3libs.log.debug(
                    "DB table name given in store URL fragment,"
                    " but overriden by `table` argument to SqlStore()")
            self.table_name = table_name

        # save ctor args for lazy-initialization
        self._init_extra_fields = (extra_fields if extra_fields is not None else {})
        self._init_create = create

        # create slots for lazy-init'ed attrs
        self._real_engine = None
        self._real_extra_fields = None
        self._real_tables = None

        self._loaded = WeakValueDictionary()

    @staticmethod
    def _to_sqlalchemy_url(url):
        if url.scheme == 'sqlite':
            # rewrite ``sqlite`` URLs to be RFC compliant, see:
            # https://github.com/uzh/gc3pie/issues/261
            db_url = "%s://%s/%s" % (url.scheme, url.netloc, url.path)
        else:
            db_url = str(url)
        # remove fragment identifier, if any
        try:
            fragment_loc = db_url.index('#')
            db_url = db_url[:fragment_loc]
        except ValueError:
            pass
        return db_url

    def _delayed_init(self):
        """
        Perform initialization tasks that can interfere with
        forking/multiprocess setup.

        See `GC3Pie issue #550 <https://github.com/uzh/gc3pie/issues/550>`_
        for more details and motivation.
        """
        url = self._to_sqlalchemy_url(self.url)
        gc3libs.log.debug(
            "Initializing SQLAlchemy engine for `%s`...", url)
        self._real_engine = sqla.create_engine(url)

        # create schema
        meta = sqla.MetaData(bind=self._real_engine)
        table = sqla.Table(
            self.table_name,
            meta,
            sqla.Column('id',
                        sqla.Integer(),
                        primary_key=True, nullable=False),
            sqla.Column('data',
                        sqla.LargeBinary()),
            sqla.Column('state',
                        sqla.String(length=128)))

        # create internal rep of table
        self._real_extra_fields = {}
        for col, func in self._init_extra_fields.items():
            assert isinstance(col, sqla.Column)
            table.append_column(col.copy())
            self._real_extra_fields[col.name] = func

        # check if the db exists and already has a 'store' table
        current_meta = sqla.MetaData(bind=self._real_engine)
        current_meta.reflect()
        if self._init_create and self.table_name not in current_meta.tables:
            meta.create_all()

        self._real_tables = meta.tables[self.table_name]


    def pre_fork(self):
        """
        Dispose current SQLAlchemy engine (if any).
        A new SQLAlchemy engine will be initialized
        upon the next interaction with a DB.

        This method only exists to allow `SessionBasedDaemon`:class:
        and similar applications that can do DB operations after
        fork()ing to continue to operate, without incurring into a
        SQLAlchemy "OperationalError: (...) could not receive data
        from server: Transport endpoint is not connected"
        """
        if self._real_engine:
            self._real_engine.dispose()
        self._real_engine = None
        self._real_extra_fields = None
        self._real_tables = None


    @property
    def _engine(self):
        if self._real_engine is None:
            self._delayed_init()
        return self._real_engine

    @property
    def _tables(self):
        if self._real_tables is None:
            self._delayed_init()
        return self._real_tables

    # FIXME: Remove once the TissueMAPS code is updated not to use this any more!
    @property
    def t_store(self):
        """
        Deprecated compatibility alias for `SqlStore._tables`
        """
        warn("`SqlStore.t_store` has been renamed to `SqlStore._tables`;"
             " please update your code", DeprecationWarning, 2)
        return self._tables

    @property
    def extra_fields(self):
        if self._real_extra_fields is None:
            self._delayed_init()
        return self._real_extra_fields


    @same_docstring_as(Store.invalidate_cache)
    def invalidate_cache(self):
        self._loaded.clear()

    @same_docstring_as(Store.list)
    def list(self):
        q = sql.select([self._tables.c.id])
        with self._engine.begin() as conn:
            rows = conn.execute(q)
            ids = [i[0] for i in rows.fetchall()]
        return ids

    @same_docstring_as(Store.replace)
    def replace(self, id_, obj):
        self._save_or_replace(id_, obj)

    # copied from FilesystemStore
    @same_docstring_as(Store.save)
    def save(self, obj):
        if not hasattr(obj, 'persistent_id'):
            obj.persistent_id = self.idfactory.new(obj)
        return self._save_or_replace(obj.persistent_id, obj)

    def _save_or_replace(self, id_, obj):
        # if __debug__:
        #     global _lvl
        #     _lvl += '>'
        #     gc3libs.log.debug("%s Saving %r@%x as %s ...", _lvl, obj, id(obj), id_)

        # build row to insert/update
        fields = {'id': id_}

        with closing(BytesIO()) as dstdata:
            make_pickler(self, dstdata, obj).dump(obj)
            fields['data'] = dstdata.getvalue()

        try:
            fields['state'] = obj.execution.state
        except AttributeError:
            # If we cannot determine the state of a task, consider it UNKNOWN.
            fields['state'] = Run.State.UNKNOWN

        # insert into db
        for column in self.extra_fields:
            try:
                fields[column] = self.extra_fields[column](obj)
            except Exception as ex:
                gc3libs.log.warning(
                    "Error saving DB column '%s' of object '%s': %s: %s",
                    column, obj, ex.__class__.__name__, str(ex))

        if __debug__:
            for column in fields:
                if column == 'data':
                    continue
                gc3libs.log.debug(
                    "Writing value '%s' in column '%s' for object '%s'",
                    fields[column], column, obj)

        q = sql.select([self._tables.c.id]).where(self._tables.c.id == id_)
        with self._engine.begin() as conn:
            r = conn.execute(q)
            if not r.fetchone():
                # It's an insert
                q = self._tables.insert().values(**fields)
            else:
                # it's an update
                q = self._tables.update().where(
                        self._tables.c.id == id_).values(**fields)
            conn.execute(q)
        obj.persistent_id = id_
        if hasattr(obj, 'changed'):
            obj.changed = False

        # update cache
        if str(id_) in self._loaded:
            old = self._loaded[str(id_)]
            if old is not obj:
                self._loaded[str(id_)] = obj
                # if __debug__:
                #     gc3libs.log.debug(
                #         "%s Overwriting object %s %r@%x with %r@%x",
                #         _lvl, id_, old, id(old), obj, id(obj))
                #     from traceback import format_stack
                #     gc3libs.log.debug("Traceback:\n%s", ''.join(format_stack()))

        # if __debug__:
        #     gc3libs.log.debug("%s Done saving %r@%x as %s ...", _lvl, obj, id(obj), id_)
        #     if _lvl:
        #         _lvl = _lvl[:-1]

        # return id
        return id_

    @same_docstring_as(Store.load)
    def load(self, id_):
        # if __debug__:
        #     global _lvl
        #     _lvl += '<'
        #     gc3libs.log.debug("%s Store %s: Loading task %s %r ...", _lvl, self, id_, type(id_))

        # return cached copy, if any
        try:
            obj = self._loaded[str(id_)]
            # if __debug__:
            #     if _lvl:
            #         _lvl = _lvl[:-1]
            #     gc3libs.log.debug("%s Store %s: Returning cached object %r@%x as task %s", _lvl, self, obj, id(obj), id_)
            return obj
        except KeyError:
            pass

        # no cached copy, load from disk
        q = sql.select([self._tables.c.data]).where(self._tables.c.id == id_)
        with self._engine.begin() as conn:
            rawdata = conn.execute(q).fetchone()
        if not rawdata:
            raise gc3libs.exceptions.LoadError(
                "Unable to find any object with ID '%s'" % id_)
        obj = make_unpickler(self, BytesIO(rawdata[0])).load()
        super(SqlStore, self)._update_to_latest_schema()
        assert str(id_) not in self._loaded
        self._loaded[str(id_)] = obj
        # if __debug__:
        #     if _lvl:
        #         _lvl = _lvl[:-1]
        #     gc3libs.log.debug("%s Store %s: Done loading task %s as %r@%x.", _lvl, self, id_, obj, id(obj))
        return obj

    @same_docstring_as(Store.remove)
    def remove(self, id_):
        with self._engine.begin() as conn:
            conn.execute(
                self._tables.delete().where(self._tables.c.id == id_))
        try:
            del self._loaded[str(id_)]
        except KeyError:
            pass


# register all URLs that SQLAlchemy can handle
def make_sqlstore(url, *args, **extra_args):
    """
    Return a `SqlStore`:class: instance, given a SQLAlchemy URL and
    optional initialization arguments.

    This function is a bridge between the generic factory functions
    provided by `gc3libs.persistence.make_store`:func: and
    `gc3libs.persistence.register`:func: and the class constructor
    `SqlStore`:class.

    Examples::

      | >>> ss1 = make_sqlstore(gc3libs.url.Url('sqlite:////tmp/foo.db'))
      | >>> ss1.__class__.__name__
      | 'SqlStore'

    """
    assert isinstance(url, gc3libs.url.Url)
    gc3libs.log.debug("Building SQL store from URL %s", url)
    if url.scheme == 'sqlite':
        # create parent directories: avoid "OperationalError: unable to open
        # database file None None"
        dir = gc3libs.utils.dirname(url.path)
        if not os.path.exists(dir):
            os.makedirs(dir)
    return SqlStore(url, *args, **extra_args)


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="sql",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gtopas.py -- Front-end script for running BEAST jobs.
#   https://www.beast2.org/
#
#   Copyright (C) 2017 2018  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple BEAST jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gtopas.py --help`` for program usage
instructions.

Input parameters consists of:
:param str BEAST XML file: Path to an .xml file containing BEAST settings 
and input data
"""

# summary of user-visible changes
__changelog__ = """
  2017-08-28:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'
__version__ = '1.0'

# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gtopas
    gtopas.GtopasScript().run()

import os
import sys
import time
import tempfile
import mimetypes
import random

import pandas
from xml.etree import cElementTree as ET
from xml.etree.ElementTree import Element, SubElement, Comment, tostring

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, positive_int, existing_file, existing_directory
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

# Utility functions
def _get_valid_input(input_folder):
    """
    Return list of valid input file in .txt format
    Input file will be returned with ABS path
    """

    return [os.path.join(input_folder,input_file) for input_file in os.listdir(input_folder) if input_file.endswith('.txt') and if mimetypes.guess_type(input_file)[0] == 'text/plain']
    

## custom application class
class GtopasApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gtopas'
    
    def __init__(self, input_file, **extra_args):

        executables = []
        inputs = dict()
        outputs = dict()
        
        Application.__init__(
            self,
            arguments = ['topas', os.path.basename(input_file_name)],
            inputs = [input_file],
            outputs = ['scores.csv'],
            # Note: aternatively, you could specify a folder name as 'output'. GC3Pie will retrieve the entire content of that folder
            # Example: outputs = ['./results/']
            stdout = 'gtopas.log',
            join=True,
            executables = executables,
            **extra_args)
    
class GtopasScript(SessionBasedScript):
    """
    The ``gtopas`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.
    
    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gtopas`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gtopas``
    aggregates them into a single larger output file located in 
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GtopasApplication, 
            stats_only_for = GtopasApplication,
            )

    def setup_args(self):        
        self.add_param('input_folder',
                       type=existing_directory,
                       help="Path to input folder containing valid input .txt files.")        

    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new GcgpsTask
        """
        tasks = []
        
        for input_file in _get_valid_input(self.params.input_folder):
            
            jobname = extract_jobname_from_file(input_file)
                
            extra_args = extra.copy()
            extra_args['jobname']  = jobname
            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)
            
            tasks.append(GtopasApplication(
                input_file,
                **extra_args))

        return tasks

    def after_main_loop(self):
        """
        Merge all result files together
        Then clean up tmp files
        """
        for task in self.session:
            if isinstance(task, GsubbeastApplication) and task.execution.returncode == 0:
                result_log_file = os.path.join(task.output_dir, '{0}.log'.format(task.id_name))
                if os.path.isfile(result_log_file):
                    gc3libs.log.debug('Reading output file {0}'.format(result_log_file))
                    data = pandas.read_csv(result_log_file, sep='\t', comment="#")
                    for key in df_dict.keys():
                        cols = [col for col in data.columns if key in col]
                        # In case multiple entries, take the first occurrence                    
                        # column_to_search = "{0}:{1}".format(key,task.id_name)
                        gc3libs.log.debug("Column [{0}] found {1} occurances. Should be 1.".format(key,len(cols)))
                        if len(cols) > 0:
                            df_dict[key][task.id_name] = data[cols[0]]
                        else:
                            gc3libs.log.error("Skipping column {0} as no occourrences have been found".format(key))

                else:
                    gc3libs.log.error('Output file {0} for task {1} not found'.format(result_log_file, task.id_name))
                    continue
            else:
                gc3libs.log.warning('Task {0} not completed or failed during execution. Ignoring'.format(task.id_name))

        gc3libs.log.info('Writing aggregated .csv results in {0}'.format(self.params.result_csv))
        for key in df_dict.keys():
            df = df_dict[key]
            df.to_csv(os.path.join(self.params.result_csv, '{0}.csv'.format(key)))

        return

    
